/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* This file implements the primarily textual displays. */

#include "conq.h"
#include "cconq.h"

#define any_borders_at(x, y, b) (aux_terrain_at(x, y, b) != 0)

#define any_connections_at(x, y, c) (aux_terrain_at(x, y, c) != 0)


/*VARARGS*/
notify(control, a1, a2, a3, a4, a5, a6, a7, a8)
char *control, *a1, *a2, *a3, *a4, *a5, *a6, *a7, *a8;
{
    char tmpnbuf[BUFSIZE];

    sprintf(tmpnbuf, control, a1, a2, a3, a4, a5, a6, a7, a8);
    /* Capitalize first char of notice. */
    if (islower(tmpnbuf[0])) tmpnbuf[0] = toupper(tmpnbuf[0]);
    sprintf(dui.text1, "%s", tmpnbuf);
    DGprintf("%s\n", dui.text1);
    /* (should add second line if needed) */
    show_toplines();
    refresh();
}

show_toplines()
{
    clear_window(dui.toplines);
    draw_text(dui.toplines, 0, 0, dui.text1);
    draw_text(dui.toplines, 0, 1, dui.text2);
}

clear_toplines()
{
    dui.text1[0] = '\0';
    dui.text2[0] = '\0';
    show_toplines();
}

char *side_name();

/* Display all the details of the currently-selected unit/hex. */

show_closeup()
{
    int u, tview, uview;
    char *filler = "Empty ";
    Unit *unit = NULL, *topunit;
    Side *side2;

    clear_window(dui.closeup);
    if (inside_area(dui.curx, dui.cury)) {
	if (terrain_visible(dui.curx, dui.cury)) {
	    tview = buildtview(terrain_at(dui.curx, dui.cury));
	} else {
	    tview = terrain_view(dside, dui.curx, dui.cury);
	}
	if (units_visible(dui.curx, dui.cury)) {
	    unit = unit_at(dui.curx, dui.cury);
	    if (unit == NULL && dui.curunit != NULL) {
		unit = dui.curunit;
	    }
	    if (unit != NULL && alive(unit)) {
		/* If there is a unit there, we can at least see basic info. */
		sprintf(tmpbuf, "%s", unit_handle(dside, unit));
		draw_text(dui.closeup, 0, 0, tmpbuf);
		if (unit->side == dside) {
		    draw_unit_details(unit);
		}
		filler = "In ";
	    }
	} else if ((uview = unit_view(dside, dui.curx, dui.cury)) != EMPTY) {
	    filler = "In ";
	    side2 = side_n(vside(uview));
	    u = vtype(uview);
	    sprintf(tmpbuf, "%s %s",
		    (side2 == NULL ? "independent" :
		     (side2 == dside ? "your" : side_adjective(side2))),
		    u_type_name(u));
	    draw_text(dui.closeup, 0, 0, tmpbuf);
	}
	/* Describe the cell here. */
	describe_cell(dui.curx, dui.cury, tview, filler, tmpbuf);
	draw_text(dui.closeup, 0, 1, tmpbuf);
    } else {
	sprintf(tmpbuf, "??? Off-area %d,%d ???", dui.curx, dui.cury);
	draw_text(dui.closeup, 0, 1, tmpbuf);
    }
}

describe_cell(x, y, tview, filler, buf)
int x, y, tview;
char *filler, *buf;
{
    int t, t2, dir;
    char *featname;

    if (tview != UNSEEN) {
	t = vterrain(tview);
	/* Now describe terrain and position. */
	sprintf(buf, "%s%s", filler, t_type_name(t));
	featname = feature_name_at(dui.curx, dui.cury);
	if (featname) {
	    sprintf(buf+strlen(buf), " %s", featname);
	}
	if (temperatures_defined()) {
	    sprintf(buf+strlen(buf), " (Temp %d)",
		    temperature_at(dui.curx, dui.cury));
	}
	if (elevations_defined()) {
	    sprintf(buf+strlen(buf), " (Elev %d)",
		    elev_at(dui.curx, dui.cury));
	}
	if (any_aux_terrain_defined()) {
	    for_all_terrain_types(t2) {
		if (t_is_border(t2)
		    && aux_terrain_defined(t2)
		    && any_borders_at(dui.curx, dui.cury, t2)) {
		    sprintf(buf+strlen(buf), " %s", t_type_name(t2)); 
		    for_all_directions(dir) {
			if (border_at(dui.curx, dui.cury, dir, t2)) {
			    sprintf(buf+strlen(buf), "/%s", dirnames[dir]);
			}
		    }
		}
		if (t_is_connection(t2)
		    && aux_terrain_defined(t2)
		    && any_connections_at(dui.curx, dui.cury, t2)) {
		    sprintf(buf+strlen(buf), " %s", t_type_name(t2)); 
		    for_all_directions(dir) {
			if (connection_at(dui.curx, dui.cury, dir, t2)) {
			    sprintf(buf+strlen(buf), "/%s", dirnames[dir]);
			}
		    }
		}
	    }
	}
    } else {
	sprintf(buf, "<unknown>");
    }
    sprintf(buf+strlen(buf), " at %d,%d", dui.curx, dui.cury);
}

/* (do we still want this?) */

draw_history()
{
#if 0
    int age, u;
    short view, prevview;
    Side *side2;

    /* Compose and display view history of this hex. */
    Dprintf("Drawing previous view info\n");
    age = side_view_age(side, dui.curx, dui.cury);
    prevview = side_prevview(side, dui.curx, dui.cury);
    if (age == 0) {
	if (prevview != view) {
	    if (prevview == EMPTY) {
		/* misleading if prevview was set during init. */
		sprintf(tmpbuf, "Up to date; had been empty.");
	    } else if (prevview == UNSEEN) {
		sprintf(tmpbuf, "Up to date; had been unexplored.");
	    } else {
		side2 = side_n(vside(prevview));
		u = vtype(prevview);
		if (side2 != side) {
		    sprintf(tmpbuf, "Up to date; had seen %s %s.",
			    (side2 == NULL ? "independent" :
			     side_name(side2)),
			    u_type_name(u));
		} else {
		    sprintf(tmpbuf,
			    "Up to date; had been occupied by your %s.",
			    u_type_name(u));
		}
	    }
	} else {
	    sprintf(tmpbuf, "Up to date.");
	}
    } else {
	if (prevview == EMPTY) {
	    sprintf(tmpbuf, "Was empty %d turns ago.", age);
	} else if (prevview == UNSEEN) {
	    sprintf(tmpbuf, "Terrain first seen %d turns ago.", age);
	} else {
	    side2 = side_n(vside(prevview));
	    u = vtype(prevview);
	    if (side2 != side) {
		sprintf(tmpbuf, "Saw %s %s, %d turns ago.",
			(side2 == NULL ? "independent" :
			 side_name(side2)),
			u_type_name(u), age);
	    } else {
		sprintf(tmpbuf, "Was occupied by your %s %d turns ago.",
			u_type_name(u), age);
	    }
	}
    }
    draw_text(dui.closeup, 0, 2, tmpbuf);
#endif
}

/* Describe the state of the given unit, in maximal detail. */

draw_unit_details(unit)
Unit *unit;
{
    int i, u = unit->type, m, nums[MAXUTYPES], xpos;
    int terr = terrain_at(unit->x, unit->y);
    Unit *occ, *top;

    /* Say which unit this is. */
    sprintf(spbuf, "%s", unit_handle(dside, unit));
    /* Describe the "important" parameters like hit points and moves. */
    sprintf(tmpbuf, "   HP %d/%d", unit->hp, u_hp(u));
    strcat(spbuf, tmpbuf);
    /* (should say something about parts here) */
    if (!completed(unit)) {
	sprintf(tmpbuf, "  %d/%d done", unit->cp, u_cp(u));
	strcat(spbuf, tmpbuf);
    } else {
	if (u_acp(u) > 0) {
	    sprintf(tmpbuf, "  Ac %d/%d", unit->act->acp, unit->act->initacp);
	    strcat(spbuf, tmpbuf);
	}
    }
    draw_text(dui.closeup, 0, 0, spbuf);
    /* Mention transport and other units stacked here. */
    if (unit->transport != NULL) {
	sprintf(spbuf, "In %s", short_unit_handle(unit->transport));
    } else {
	describe_cell(unit->x, unit->y, buildtview(terr), "In ", spbuf);
	/* make this a "stacked_at" macro? */
	if (((top = unit_at(unit->x, unit->y)) != NULL)
	    && top->nexthere != NULL) {
	    strcat(spbuf, ", ");
	    for_all_unit_types(i) nums[i] = 0;
	    for_all_stack(unit->x, unit->y, occ) ++nums[occ->type];
	    --nums[u];  /* don't count ourself */
	    for_all_unit_types(i) {
		if (nums[i] > 0) {
		    sprintf(tmpbuf, "%d %1s  ", nums[i], utype_name_n(i, 1));
		    strcat(spbuf, tmpbuf);
		}
	    }
	    strcat(spbuf, " here also");
	}
    }
    draw_text(dui.closeup, 0, 1, spbuf);
    /* Very briefly list the numbers and types of the occupants. */
    sprintf(spbuf, "");
    if (unit->occupant != NULL) {
	strcpy(spbuf, "Occ ");
	for_all_unit_types(i) nums[i] = 0;
	for_all_occupants(unit, occ) ++nums[occ->type];
	for_all_unit_types(i) {
	    if (nums[i] > 0) {
		sprintf(tmpbuf, "%d %1s  ", nums[i], utype_name_n(i, 1));
		strcat(spbuf, tmpbuf);
	    }
	}
    }
    draw_text(dui.closeup, 0, 2, spbuf);
#if 0
    /* Describe the state of construction. */
    if (building(unit)) {
	sprintf(spbuf, "Working on a %s",
		u_type_name(unit->plan->abuilding->type));
	/* (should also describe other building plans? or put into
	   plan description instead?) */
	draw_text(dui.closeup, 40, 1, spbuf);
    }
#endif
    /* Describe the state of all the supplies. */
    sprintf(spbuf, "");
    for_all_material_types(m) {
	if (um_storage(u, m) > 0) {
	    sprintf(tmpbuf, "%s %d/%d  ",
		    m_type_name(m), unit->supply[m], um_storage(u, m));
	    strcat(spbuf, tmpbuf);
	}
    }
    draw_text(dui.closeup, 0, 3, spbuf);
    /* Describe the current plans, tasks, etc. */
    /* (needs much improvement) */
    if (unit->plan) {
	char *pd;

	pd = plan_desig(unit->plan);
	draw_text(dui.closeup, 0, 4, pd);
    }
}

show_side_closeup()
{
    clear_window(dui.sidecloseup);
    draw_side_closeup(dui.sidecloseup, dui.side);
}

/* Don't mix up the side you're looking at and the side you're displaying on */

draw_side_closeup(sidecloseup, side2)
Side *side2;
int sidecloseup;
{
    int line = 0, u;

    if (side2->name) {
	sprintf(spbuf, "%s", side2->name);
    } else if (side2->adjective) {
	sprintf(spbuf, "The %s side", side2->adjective);
    } else {
	sprintf(spbuf, "unnamed side(?)");
    }
    if (side2->lost) {
	sprintf(spbuf+strlen(spbuf), " (lost)");
    }
    draw_text(sidecloseup, 0, (line++), spbuf);
/*    sprintf(spbuf, "Strategy is under %s control.",
	    (side->strategy ? "automatic" :
	     (side == side2 ? "your" : "human"))); */
    draw_text(sidecloseup, 0, (line++), spbuf);
    sprintf(spbuf, "Ask Side %d, Time Out %d Seconds",
	    side2->unitseveraskside, side2->realtimeout);
    draw_text(sidecloseup, 0, (line++), spbuf);
    for_all_unit_types(u) {
	sprintf(spbuf, "%s: %s ask side, %s bad terrain,",
		u_type_name(u),
		(u_doctrine(side2, u, everaskside) ? "may" : "never"),
		(u_doctrine(side2, u, avoidbadterrain) ? "avoid" : "ignore")
		);
	draw_text(sidecloseup, 0, (line++), spbuf);
	sprintf(spbuf, "    build %s.",
		(u_doctrine(side2, u, buildalways) ? "always" : "sometimes"));
	draw_text(sidecloseup, 0, (line++), spbuf);
    }
}

/* Basic routine that displays the list of sides. */

show_side_list()
{
    int sy = 0, i, totacp;
    char progress[BUFSIZE];
    Side *side2;

    /* should ensure subwin is clear */
    clear_window(dui.sides);
    for_all_sides(side2) {
	if (side2->ingame) {
	    totacp = side_initacp(side2);
	    if (totacp > 0) {
		sprintf(progress, "%3d%%", (100 * side_acp(side2)) / totacp);
		/* We get to see our actual total acp as well. */
		if (dside == side2) {
		    sprintf(progress+strlen(progress), "/%3d", totacp);
		}
	    } else {
		sprintf(progress, "----");
	    }
	} else {
	    if (side2->lost) {
		sprintf(progress, "Lost");
	    } else if (side_won(side2)) {
		sprintf(progress, "Won!");
	    } else {
		sprintf(progress, "Gone");
	    }
	}
	sprintf(spbuf, "%s%d%c %s %s (%s)",
		(side2->designer ? "!!!" : ""),
		side_number(side2),
		(side2->finishedturn ? ' ' : '*'),
		progress,
		side_name(side2),
		(side2->player->displayname ?
		 side2->player->displayname : ""));
	draw_text(dui.sides, 0, sy, spbuf);
	sy += 1;
    }
    /* Draw the modeline. */
    standout();
    draw_text(dui.sides, 0, dui.sh - 1,
	      "--Sides-------------------------------");
    standend();
}

/* Update the time and optional season display. */

show_game_date()
{
    clear_window(dui.date);
    /* First line of the game state. */
    sprintf(spbuf, "%s", absolute_date_string(curturn));
    if (world.yearlength > 1) {
	sprintf(tmpbuf, " (%s)", season_name(curturn));
	strcat(spbuf, tmpbuf);
    }
    draw_text(dui.date, 0, 0, spbuf);
    refresh();
}

/* General list display routine. */

/* (should track beginning/end of displayed list, draw only visible elts) */
/* (should add scrolling interaction) */

show_list()
{
    int i = 0, line = 1;
    char *sortorder = "???", *sideslisted = "xxx";
    Unit *unit;
    Side *loopside, *side2;

    clear_window(dui.list);
    switch (dui.listsides) {
      case ourside:
	sideslisted = "your side";
	break;
      case ourallies:
	sideslisted = "yours and allied sides";
	break;
      case allsides:
	sideslisted = "all sides";
	break;
    }
    switch (dui.listtype) {
      case byrandorder:
	sortorder = "random";
	switch (dui.listsides) {
	  case ourside:
	    for_all_side_units(dside, unit) {
		line += draw_unit_list_entry(unit, line);
	    }
	    break;
	  case ourallies:
	    for_all_units(loopside, unit) {
		if (allied_side(dside, loopside)) {
		    line += draw_unit_list_entry(unit, line);
		}
	    }
	    break;
	  case allsides:
	    for_all_units(loopside, unit) {
		if (unit->side == dside || 0 /* covered */) {
		    line += draw_unit_list_entry(unit, line);
		}
	    }
	    break;
	}
	break;
      case bytype:
	sortorder = "type";
	switch (dui.listsides) {
	  case ourside:
	    break;
	  case ourallies:
	    break;
	  case allsides:
	    break;
	}
	break;
      case byactorder:
	sortorder = "acting";
	switch (dui.listsides) {
	  case ourside:
	    for (unit = dside->actorder; unit != NULL; unit = unit->nextactor) {
		line += draw_unit_list_entry(unit, line);
	    }
	    break;
	  case ourallies:
	    break;
	  case allsides:
	    for_all_sides(side2) {
		for (unit = side2->actorder; unit != NULL; unit = unit->nextactor) {
		    line += draw_unit_list_entry(unit, line);
		}
	    }
	    break;
	}
	break;
      case bylocation:
	sortorder = "location";
	/* iterate over world coords */
	switch (dui.listsides) {
	  case ourside:
	    break;
	  case ourallies:
	    break;
	  case allsides:
	    break;
	}
	break;
    }
    /* Draw the modeline, in standout if possible. */
    standout();
    sprintf(spbuf, "--%s--%s--",
	    sideslisted, sortorder); 
    draw_text(dui.list, 0, dui.lh - 1, spbuf);
    standend();
    refresh();
}

/* Alter the numbers for a single type of unit.  Should be called right
   after any changes.  Formatted to look nice, but kind of messy to set
   up correctly; display should not jump back and forth as the numbers
   change in size. */

show_unit_type_list(u)
int u;
{
    if (active_display(dside)) {
	/* Tack all the bits of info onto the end of the buffer. */
	sprintf(spbuf, "");
#if 0
	/* Our unit total (right-justified) */
	if (side->units[u] > 0)	{
	    sprintf(tmpbuf, "%4d", side->units[u]);
	} else {
	    sprintf(tmpbuf, "    ");
	}
	strcat(spbuf, tmpbuf);
	/* Our units under construction (left-justified) */
	if (side->unitsbuilding[u] > 0) {
	    sprintf(tmpbuf, "(%d)", side->unitsbuilding[u]);
	} else {
	    sprintf(tmpbuf, "    ");
	}
#endif
	strcat(spbuf, tmpbuf);
#if 0
	/* Our total possessions over the game */
	/* Make sure that subsequent write goes into valid string area. */
	strcat(spbuf,  "            ");
	if (total_gain(side, u) > 0) {
	    sprintf(tmpbuf, "%4d", total_gain(side, u));
	    sprintf(spbuf+10, "%s", tmpbuf);
	}
	/* Our total losses over the game */
	strcat(spbuf,  "            ");
	if (total_loss(side, u) > 0) {
	    sprintf(tmpbuf, "- %d", total_loss(side, u));
	    sprintf(spbuf+15, "%s", tmpbuf);
	}
#endif
	draw_text(dui.list, 0, u + 2, spbuf);
    }
}

/* Would be faster to stash these, but enough difference to care? */

total_gain(side, u)
Side *side;
int u;
{
    int i, total = 0;

/*    for (i = 0; i < DUMMYREAS; ++i)
      total += side_balance(side, u, i);  */
    return total;
}

total_loss(side, u)
Side *side;
int u;
{
    int i, total = 0;

/*    for (i = DUMMYREAS+1; i < NUMREASONS; ++i)
      total += side_balance(side, u, i); */
    return total;
}

draw_unit_list_entry(unit, line)
Unit *unit;
int line;
{
    if (unit == dui.curunit) {
	draw_text(dui.list, 0, line, "*");
    }
    if (alive(unit)) {
	sprintf(spbuf, "%-30s",	unit_handle(dside, unit));
    } else {
	sprintf(spbuf, "---");
    }
    draw_text(dui.list, 1, line, spbuf);
    return 1;
}

show_clock()
{
    int time = 0 /* time remaining? */;
    int hour, minute, second;

    if (realtime_game()) {
	hour = time / 3600;  minute = (time / 60) % 60;  second = time % 60;
	sprintf(spbuf, "%s%d:%s%d:%s%d",
		(hour < 10 ? "0" : ""), hour,
		(minute < 10 ? "0" : ""), minute,
		(second < 10 ? "0" : ""), second);
	draw_text(dui.clock, 0, 0, spbuf);
    }
}
