(game-module "napoleon"
  (title "Napoleon")
  (blurb "definitions for the Napoleonic wars")
  (variants (see-all) (world-seen) (world-size))
)

(unit-type i (name "infantry") (image-name "soldiers")
  (help "the backbone of the army"))
(unit-type g (name "guards") (image-name "soldiers")
  (help "the elite of the army"))
(unit-type c (name "cavalry") (image-name "cavalry")
  (help "zips around for shock effect"))
(unit-type ldr (name "leader") (image-name "flag")
  (help "controls the movements of land forces"))
(unit-type t (name "transport") (image-name "brig")
  (help "how armies get across water"))
(unit-type f (name "frigate") (image-name "frigate")
  (help "the eyes of the fleet"))
(unit-type fleet (image-name "twodecker")
  (help "a full line of battle fleet"))
(unit-type balloon (image-name "balloon")
  (help "fragile but good for reconnaissance"))
(unit-type / (name "entrenchments") (image-name "camp")
  (help "temporary protection for armies"))
(unit-type city (name "city") (image-name "town20")
  (help "a typical city"))
(unit-type capital (name "capital") (image-name "city18")
  (help "the major city of a country"))

(material-type food)
(material-type shot)

(include "stdterr")  ; should simplify probably

(define water (sea shallows))
(define land (plains desert forest mountains))
(define movers (i g c ldr t f fleet balloon))
(define land-forces (i g c))
(define ships (t f fleet))
(define places (/ city capital))

(add places point-value (0 5 25))

(add u* already-seen true)

;;; Static relationships.

;;; Unit-unit capacities.

(table unit-size-as-occupant
  (u* u* 100) ; disables occupancy usually
  (land-forces (t fleet) 1)
  (ldr u* 1)
  (movers (city capital) 1)
  )

(add (t fleet / city capital) capacity (4 2 10 100 100))

(table occupant-max
  (u* u* 100)
  )

;;; Unit-terrain interaction.

(table vanishes-on
  (land-forces sea true)
  (ships land true)
  )

;;; Unit-terrain capacities.

;; Allow effectively infinite capacity everywhere.

(table unit-size-in-terrain
  (u* t* 1)
  )

(add t* capacity 100)

;;; Action parameters.

(add u* acp-per-turn (8 10 3 10 2 5 3 5 0 1 1))
;; capitals should be more capable?

;;; Movement.

(add movers speed    (8 10 3 10 2 5 3 5))

(add movers acp-per-move 100)

(table mp-to-enter-terrain
  (u* t* 1)
  (land-forces sea 20)
  (c (forest mountains) 2)
  ;; Leaders should move about as fast as a frigate, when at sea.
  (ldr (sea shallows) 2)
  ;; Leaders are also slower in the mountains.
  (ldr (mountains) 2)
  (ships land 20)
  ;; add any effect for shallows?
  )

;;; Construction.

(add u* cp (20 40 8 12 10 20 40 4 1 1 1))

(table acp-to-create
  ((i c) / 1)
  ((city capital) movers 1)
  (city balloon 0)  ; only let capitals build balloons
  )

(table cp-on-creation
  ((i c) / 1)
  ((city capital) movers 1)
)

(table acp-to-build
  ((city capital) movers 1)
  (city balloon 0)  ; only let capitals build balloons
  )

(table cp-per-build
  ((city capital) movers 1)
  )

;;; (ships should need lots of tooling up)

;1 food land-forces produce
;1 food / produce
;10 [ food shot ] [ city capital ] produce
;100 [ plains forest ] land-forces productivity
;100 [ plains forest mountains ] / productivity
;0 desert / productivity
;100 plains [ city capital ] productivity
;70 [ forest mountains ] [ city capital ] productivity
;30 desert [ city capital ] productivity
;[ 4 10  4 300 300 300  0 10 500 1000 ] food u* storage
;[ 1  0 30  30  30  50  0 100 200 400 ] shot u* storage
;1 food land-forces consume
;1 food ships consume
;-1 r* land-forces out-length

;;; Combat.

(add u* hp-max (10 10 2 1 3 6 1 1 10 20 40))

(add (i g) parts-max 10)

(table hit-chance
  (u* u* 50)
  ;; Leaders should never expose themselves to combat directly.
  (u* ldr 90)
  (ldr u* 0)
  )

;[ 40  50  30  20  10   5  75  99  99  99 ] u* i hit
;[ 40  50  30  10   5   2  50  50  50  50 ] u* c hit
;[ 60  70  60  50  50  40  20  99  99  99 ] u* a hit
;[ 20  30  10  20  10  10  10  10  10  10 ] u* t hit
;[ 30  40  20  60  50  50  20  60  60  60 ] u* f hit
;[ 40  50  30  80  40  50  50 100 100 100 ] u* fleet hit
;[ 10  10  10  10  10  10  10  10  10  10 ] u* balloon hit
;[  0   0   0   0   0   0   0   0   0   0 ] u* / hit
;[ 20  20  20  20  20  20  20  20  20  20 ] u* * hit
;[ 40  40  40  40  40  40  40  40  40  40 ] u* capital hit

(table damage
  (u* u* 1)
  (fleet u* 3)
  )

(table capture-chance
  ((i g) city (20 30))
  ((i g) capital (10 15))
  )

;1 ships can-disband

;1 u* [ * capital ] repair
;10 [ f fleet ] [ f fleet ] repair

(side-library set
  ((name "France") (adjective "French"))
  ((name "England") (adjective "English"))
  ((name "Spain") (adjective "Spanish"))
  ((name "Austria") (adjective "Austrian"))
  ((name "Prussia") (adjective "Prussian"))
  ((name "Russia") (adjective "Russian"))
  )

(world 1000 (year-length 12))

;;; Random setup stuff.  Sort of meaningless for a historical game, but useful.

(add t* alt-percentile-min (  0  50  50  51  51  51  95 0 0 0))
(add t* alt-percentile-max ( 50  51  51  95  95  95 100 0 0 0))
(add t* wet-percentile-min (  0   0  50   0  10  90   0 0 0 0))
(add t* wet-percentile-max (100 100 100  10  90 100 100 0 0 0))

(add plains country-terrain-min 7)
(set country-radius-min 3)
(set country-separation-min 8)

(add (city capital) start-with (5 1))

;(add city independent-density 1000)

(add table favored-terrain
  (u* t* 0)
  ((city capital) plains 100)
  )

(scorekeeper (do last-side-wins))

(game-module (notes
  ))

(game-module (design-notes
  "Map scale is 50 km/hex, game time is 1 month/turn."
  ""
  "Balloons are more fun than realistic."
  ))
