(game-module "ww2-adv"
  (title "Advanced WWII Defns")
  (blurb "Types for global-scale simulation of WWII, at national level")
)

(unit-type infantry (image-name "infantry")
  (help "a primarily foot-powered army"))
(unit-type armor (image-name "armor")
  (help "a primarily mechanized army"))

(define land-units (infantry armor))

(unit-type air-force
  (help "strategic bombing capability"))
(unit-type bomber
  (help "strategic bombing capability"))
(unit-type interceptor
  (help "anti-bomber capability"))

(define air-types (air-force bomber interceptor))

(unit-type convoy (image-name "ap"))
(unit-type fleet (image-name "bb"))
(unit-type cv-fleet (image-name "cv"))
(unit-type sub-fleet (image-name "sub"))
(unit-type asw-fleet (image-name "asw"))

(define ship-types (convoy fleet cv-fleet sub-fleet asw-fleet))

(unit-type city (image-name "city20")
  (help "major cities only"))
(unit-type port-city (image-name "port")
  (help "major port cities only"))
(unit-type capital (image-name "city20")
  (help "capital of a country or region"))
(unit-type port-capital (image-name "port")
  (help "capital that is also a port"))

(define places (city port-city capital port-capital))

(define inf infantry)

(material-type oil)

(terrain-type sea)
(terrain-type land (image-name "plains"))
(terrain-type mountains)
(terrain-type non-beach (subtype border))
(terrain-type river (subtype border))

(define land-types (land mountains))

;;; Static relationships.

(table vanishes-on
  (land-units sea true)
  (ship-types land-types true)
  (places sea true)
  )

;;; Actions.

(add u* acp-per-turn 1)

(add land-units acp-per-turn (4 5))

;;; Movement.

(add u* speed 1)
(add land-units speed (4 5))
(add places speed 0)

;;; Combat.

;;; The world.

(world 360 (year-length 12))

; Area should go from 70 North to 50 South, perhaps do some distortion
; to shrink world.

(side-library set
  ((name "America") (short-name "USA"))
  ((name "England") (short-name "UK"))
  ((name "Russia"))
  ((name "France"))
  ((name "Germany"))
  ((name "Italy"))
  ((name "Japan"))
  ;; and other minor countries...
  )

;;; Setup parameters for testing.

(add t* alt-percentile-min (  0  60  95 0 0))
(add t* alt-percentile-max ( 60  95 100 0 0))
(add t* wet-percentile-min (  0   0   0 0 0))
(add t* wet-percentile-max (100 100 100 0 0))

(set country-radius-min 4)

(add u* start-with 1)

(add (sea land) country-terrain-min (4 4))

(table favored-terrain
  (u* t* 0)
  (land-units land 100)
  (air-types land 100)
  (ship-types sea 100)
  (places land 100)
  )

(add land country-people-chance 100)

(game-module (design-notes
  "The basic idea of this game is to model the grand strategy of the WWII era."
  "Scale is 100 km/hex, 1 month/turn."
  ""
  "At this scale the details of maneuver become less important than managing"
  "production and logistics."
  ""
  "Land units are infantry and armor armies, basically the same except that armor"
  "is more highly mechanized."
  ""
  "Air forces represent strategic bombing abilities."
  ""
  "Convoys are primarily transport ships with some protection."
  ""
  "Fleets are battleship/cruiser fleets, while carrier fleets have a long-range"
  "strike capability due to their aircraft."
  "Submarine fleets represent a large number of subs operating over a wide area."
  ""
  "City production is very hard to change."
  ))

