/* Copyright (c) 1992-1993  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Map interaction for the Mac interface. */

#include "conq.h"
#include "mac.h"

extern int mayseeall;

extern int weseeall;

extern RGBColor graycolor, blackcolor;

extern int conwid, tophgt;

#define terrain_visible(x, y)  \
  (g_see_all() || dside->designer || terrain_view(dside, wrapx(x), y) != UNSEEN)

#define units_visible(x, y)  \
  (g_see_all() || dside->designer || cover(dside, wrap(x), y) >= 1)

/* This is a temporary used by procs. */

Map *curmap = NULL;

/* This scroll proc is shared by both the horizontal and vertical scrollbars. */

pascal void
map_scroll_proc(control, code)
ControlHandle control;
short code;
{
	int curvalue, pagesize, jump;

	if (curmap == NULL) return;
	/* The page jump should be most but not all of a screenful. */
	if (control == curmap->hscrollbar) {
		pagesize = (2 * curmap->pxw) / 3;
	} else {
		pagesize = (2 * curmap->pxh) / 3;
	}
	switch (code) {
		case inPageDown:
			jump = pagesize;
			break;
		case inDownButton:
			jump = 4;
			break;
		case inPageUp:
			jump = 0 - pagesize;
			break;
		case inUpButton:
			jump = -4;
			break;
		default:
			jump = 0;
			break;
	}
	/* Letting control's code do the max/min hacking. */
	SetCtlValue(control, GetCtlValue(control) + jump);
	curvalue = GetCtlValue(control);
	/* Tweak the map's own variables to match. */
	if (control == curmap->hscrollbar) {
		curmap->sx = curvalue;
	} else {
		curmap->sy = curvalue;
	}
	curmap = NULL;
}

/* Handle a mouse down in the map window. */

do_mouse_down_map(map, mouse, mods)
Map *map;
Point mouse;
int mods;
{
	ControlHandle control;
	short part, value;
	long oldsx, oldsy;
	WindowPtr window = map->window;

	part = FindControl(mouse, window, &control);
	if (control == map->hscrollbar) {
		/* Handle the horizontal scrollbar. */
		oldsx = map->sx;
		part = FindControl(mouse, window, &control);
		switch (part) {
			case inThumb:
				/* (should add tracking ability) */
				part = TrackControl(control, mouse, NULL);
				map->sx = GetCtlValue(control);
				break;
			default:
				curmap = map;
				part = TrackControl(control, mouse, (ProcPtr) map_scroll_proc);
				break;
		}
		if (oldsx != map->sx) {
			focus_on_center(map);
			force_map_update(map);
			draw_related_maps(map);
		}
	} else if (control == map->vscrollbar) {
		/* Handle the vertical scrollbar. */
		oldsy = map->sy;
		part = FindControl(mouse, window, &control);
		switch (part) {
			case inThumb:
				part = TrackControl(control, mouse, NULL);
				map->sy = GetCtlValue(control);
				break;
			default:
				curmap = map;
				part = TrackControl(control, mouse, (ProcPtr) map_scroll_proc);
				break;
		}
		if (oldsy != map->sy) {
			focus_on_center(map);
			force_map_update(map);
			draw_related_maps(map);
		}
	} else if (mouse.h <= conwid) {
		/* Interpret as a control panel hit. */
		do_mouse_down_map_control_panel(map, mouse.h, mouse.v, mods);
	} else {
		do_mouse_down_map_content(map, mouse.h, mouse.v, mods);
	}
}

do_mouse_down_map_control_panel(map, h, v, mods)
Map *map;
int h, v, mods;
{
	int winh = map->window->portRect.bottom - map->window->portRect.top;

	/* (should better organize tests here) */
	if (between(winh - 2 * sbarwid, v, winh)) {
		switch ((winh - v) / sbarwid) {
			case 0:
				zoom_map(map, ((h < conwid / 2) ? -1 : 1));
				break;
			case 1:
				modaltool = 1;
				break;
		}
	} else if (between(0, v, 100)) {
		switch (v / 20) {
			case 0:
				if (h < conwid / 2) {
					select_previous_mover(map);
				} else {
					select_next_mover(map);
				}
				break;
			case 1:
				if (h < conwid / 2) {
					select_previous_actor(map);
				} else {
					select_next_actor(map);
				}
				break;
			case 2:
				if (h < conwid / 2) {
					select_previous_unit(map);
				} else {
					select_next_unit(map);
				}
				break;
			case 3:
				break;
			case 4:
				break;
		}
	} else if (mayseeall) {
		weseeall = !weseeall;
		force_map_update(map);
	}
}

zoom_map(map, inout)
Map *map;
int inout;
{
	int newpower = map->power + inout;

	if (newpower > NUMPOWERS - 1) newpower = NUMPOWERS-1;
	if (newpower < 0) newpower = 0;
	if (map->power != newpower) {
		set_map_power(map, newpower);
		center_on_focus(map);
		set_map_viewport(map);
		set_map_scrollbars(map);
		force_map_update(map);
	}
}

set_map_mag(map, newpower)
Map *map;
int newpower;
{
	if (newpower > NUMPOWERS - 1) newpower = NUMPOWERS-1;
	if (newpower < 0) newpower = 0;
	if (map->power != newpower) {
		set_map_power(map, newpower);
		center_on_focus(map);
		set_map_viewport(map);
		set_map_scrollbars(map);
		force_map_update(map);
	}
}

toggle_map_grid(map)
Map *map;
{
	map->drawgrid = !map->drawgrid;
	/* (should not do a total redraw?) */
	force_map_update(map);
}

toggle_map_topline(map)
Map *map;
{
	Rect tmprect;
	GrafPtr oldport;
	RgnHandle tmprgn;

	map->toph = (map->toph ? 0 : tophgt);
	set_content_rect(map);
	if (map->toph) {
 		GetPort(&oldport);
		SetPort(map->window);
#if 0
		tmprect = map->window->portRect;
		tmprect.left += conwid;
		tmprect.right -= sbarwid;  tmprect.bottom -= sbarwid;
/*		ScrollRect(&tmprect, 0, map->toph, tmprgn); */
		/* (should draw the topline here now) */
#endif
		SetPort(oldport);
		force_map_update(map);
	} else {
		force_map_update(map);
	}
}

toggle_map_other_maps(map)
Map *map;
{
	map->drawothermaps = !map->drawothermaps;
	/* (should not do a total redraw) */
	force_map_update(map);
}

toggle_map_lighting(map)
Map *map;
{
	map->drawlighting = !map->drawlighting;
	/* We have to do a total redraw. */
	force_map_update(map);
}

toggle_map_names(map)
Map *map;
{
	map->drawnames = !map->drawnames;
	/* (if now on, should draw names on top of everything, don't redraw everything) */
	force_map_update(map);
}

toggle_map_people(map)
Map *map;
{
	map->drawpeople = !map->drawpeople;
	force_map_update(map);
}

toggle_map_elevations(map)
Map *map;
{
	map->drawelevations = !map->drawelevations;
	force_map_update(map);
}

toggle_map_materials(map, m)
Map *map;
int m;
{
	map->drawmaterials[m] = !map->drawmaterials[m];
	force_map_update(map);
}

static int selrect;
static int downx, downy, downdir;

do_mouse_down_map_content(map, h, v, mods)
Map *map;
int h, v, mods;
{
	int i;
	Unit *unit;
	
	/* Remember this hex. */
	nearest_cell(map, h, v, &downx, &downy);
	/* Assume that last place clicked is a reasonable focus. */
	if (inside_area(downx, downy)) {
		map->vcx = downx;  map->vcy = downy;
	}
	if (modaltool) {
		if (modaltool == 1) {
			select_area_and_zoom(map, h, v, mods);
		}
	} else if (mods & cmdKey) {
		for (i = 0; i < map->numselections; ++i) {
			if ((unit = map->selections[i]) != NULL) {
				move_the_selected_unit(map, unit, h, v);
			}
		}
	} else if (mods & optionKey) {
		for (i = 0; i < map->numselections; ++i) {
			if ((unit = map->selections[i]) != NULL) {
				fire_the_selected_unit(map, unit, h, v);
			}
		}
	} else if (mods & shiftKey) {
		nearest_unit(map, h, v, &unit);
		if (unit && side_sees_unit(dside, unit)) {
			/* Invert the selection status of the unit. */
			if (unit_is_selected(map, unit)) {
				unselect_unit_on_map(map, unit);
				erase_selection(map, unit);
			} else {
				select_unit_on_map(map, unit);
				draw_selections_at(map, unit->x, unit->y);
			}
		} else {
			select_all_dragged_over(map, h, v, mods);
		}
	} else {
		/* Interpret an unmodified mouse down. */
#ifdef DESIGNERS
		if (dside->designer && tooltype != notool) {
			apply_designer_tool(map, h, v, mods);
		} else
#endif /* DESIGNERS */
		if (moveonclick) {
			/* Usually will only be one to move, but be general anyway. */
			for (i = 0; i < map->numselections; ++i) {
				if ((unit = map->selections[i]) != NULL) {
					move_the_selected_unit(map, unit, h, v);
				}
			}
		} else {
			unselect_all(map);
			nearest_unit(map, h, v, &unit);
			if (unit && side_controls_unit(dside, unit)) {
				select_unit_on_map(map, unit);
				draw_selections(map);
				move_on_drag(map, unit, h, v, mods);
			} else {
				select_all_dragged_over(map, h, v, mods);
			}
		}
	}
}

select_all_dragged_over(map, h0, v0, mods)
Map *map;
int h0, v0, mods;
{
	Point pt0, pt1, newmouse;
	int h1, v1, drawn = FALSE, x, y;
	Rect tmprect;

	SetPt(&pt0, h0, v0);
	SetPt(&pt1, h0, v0);
	SetRect(&tmprect, h0, v0, h0, v0);
	/* (should be a generic subr?) */
	PenMode(patXor);
	PenPat(qd.gray);
	while (WaitMouseUp()) {
		GetMouse(&newmouse);
		if (!EqualPt(pt1, newmouse) /* && PtInRect(newmouse, &(map->window->portRect)) */) {
			if (drawn) {
				tmprect.left = min(pt0.h, pt1.h);  tmprect.top = min(pt0.v, pt1.v);
				tmprect.right = max(pt0.h, pt1.h);  tmprect.bottom = max(pt0.v, pt1.v);
				FrameRect(&tmprect);
			}
			pt1 = newmouse;
			tmprect.left = min(pt0.h, pt1.h);  tmprect.top = min(pt0.v, pt1.v);
			tmprect.right = max(pt0.h, pt1.h);  tmprect.bottom = max(pt0.v, pt1.v);
			FrameRect(&tmprect);
			drawn = TRUE;
		}
	}
	if (drawn) {
		tmprect.left = min(pt0.h, pt1.h);  tmprect.top = min(pt0.v, pt1.v);
		tmprect.right = max(pt0.h, pt1.h);  tmprect.bottom = max(pt0.v, pt1.v);
		FrameRect(&tmprect);
	}
	PenNormal();
	select_all_units_in_rect(map, &tmprect);
}

select_area_and_zoom(map, h0, v0, mods)
Map *map;
int h0, v0, mods;
{
	Point pt0, pt1, newmouse;
	int drawn = FALSE, x, y;
	Rect tmprect;

	SetPt(&pt0, h0, v0);
	SetPt(&pt1, h0, v0);
	/* (should be a generic subr) */
	PenMode(patXor);
	PenPat(qd.gray);
	while (WaitMouseUp()) {
		GetMouse(&newmouse);
		if (!EqualPt(pt1, newmouse) /* && PtInRect(newmouse, &(map->window->portRect)) */) {
			if (drawn) {
				tmprect.left = min(pt0.h, pt1.h);  tmprect.top = min(pt0.v, pt1.v);
				tmprect.right = max(pt0.h, pt1.h);  tmprect.bottom = max(pt0.v, pt1.v);
				FrameRect(&tmprect);
			}
			pt1 = newmouse;
			tmprect.left = min(pt0.h, pt1.h);  tmprect.top = min(pt0.v, pt1.v);
			tmprect.right = max(pt0.h, pt1.h);  tmprect.bottom = max(pt0.v, pt1.v);
			FrameRect(&tmprect);
			drawn = TRUE;
		}
	}
	if (drawn) {
		tmprect.left = min(pt0.h, pt1.h);  tmprect.top = min(pt0.v, pt1.v);
		tmprect.right = max(pt0.h, pt1.h);  tmprect.bottom = max(pt0.v, pt1.v);
		FrameRect(&tmprect);
	}
	PenNormal();
	/* Done being modal. */
	modaltool = 0;
	nearest_cell(map, pt1.h, pt1.v, &x, &y);
	if (x != downx && y != downy) {
		zoom_to_fit(map, downx, downy, x, y);
	}
}

move_on_drag(map, unit, h0, v0, mods)
Map *map;
Unit *unit;
int h0, v0, mods;
{
	Point pt0, pt1, newmouse;
	int drawn = FALSE, x, y;

	SetPt(&pt0, h0, v0);
	SetPt(&pt1, h0, v0);
	/* (should be a generic subr?) */
	PenMode(patXor);
	while (WaitMouseUp()) {
		GetMouse(&newmouse);
		/* should scroll, then abort if we drag outside the window */
		if (0 /* PtInRect(newmouse, &(map->window->portRect)) */) {
		}
		if (!EqualPt(pt1, newmouse)) {
			if (drawn) {
				MoveTo(h0, v0);  LineTo(pt1.h, pt1.v);
			}
			pt1 = newmouse;
			MoveTo(h0, v0);  LineTo(pt1.h, pt1.v);
			drawn = TRUE;
		}
	}
	/* Erase the last drawn line. */
	if (drawn) {
		MoveTo(h0, v0);  LineTo(pt1.h, pt1.v);
	}
	PenNormal();
	nearest_cell(map, pt1.h, pt1.v, &x, &y);
	if (x != downx || y != downy) {
		move_the_selected_unit(map, unit, pt1.h, pt1.v);
	}
}

unselect_all(map)
Map *map;
{
	if (map->numselections > 0) {
		erase_selections(map);
		map->numselections = 0;
	}
}

select_unit_on_map(map, unit)
Map *map;
Unit *unit;
{
	if (map->numselections >= MAXSELECTIONS) return;
	map->selections[map->numselections++] = unit;
}

unit_is_selected(map, unit)
Map *map;
Unit *unit;
{
	int i;

	for (i = 0; i < map->numselections; ++i) {
		if (map->selections[i] == unit) return TRUE;
	}
	return FALSE;
}

unselect_unit_on_map(map, unit)
Map *map;
Unit *unit;
{
	int i, j;

	for (i = 0; i < map->numselections; ++i) {
		if (map->selections[i] == unit) {
			/* Keep selection list contiguous. */
			for (j = i + 1; j < map->numselections; ++j) {
				map->selections[j - 1] = map->selections[j];
			}
			--map->numselections;
			return;
		}
	}
}

erase_map_selection(map)
Map *map;
{
}

/* Given a map and a rectangle in it, select all the units whose images touch on
   that rectangle. */

select_all_units_in_rect(map, rectptr)
Map *map;
Rect *rectptr;
{
	int rectissmall = FALSE;
	int sx, sy, sw, sh;
	Side *side;
	Unit *unit;
	Rect unitrect, tmprect;
	
	/* First see if we're selecting over a large area or within a single cell. */
	if (rectptr->right - rectptr->left < map->hw
		&& rectptr->bottom - rectptr->top < map->hh) rectissmall = TRUE;
	/* Now look at all the plausible units and see if any's image intersects the rect. */
	for_all_units(side, unit) {
		if (in_play(unit)
			&& side_controls_unit(dside, unit)
			&& (rectissmall || unit->transport == NULL)) {
			xform_unit_self(map, unit, &sx, &sy, &sw, &sh);
			SetRect(&unitrect, sx, sy, sx + sw, sy + sh);
			if (SectRect(&unitrect, rectptr, &tmprect)) {
				select_unit_on_map(map, unit);
				/* (could do setport etc once...) */
				draw_selected_unit_setport(map, unit);
			}
		}
	}
}

move_the_selected_unit(map, unit, h, v)
Map *map;
Unit *unit;
int h, v;
{
	int x, y;
	Unit *other;

	nearest_cell(map, h, v, &x, &y);
	if (x != unit->x || y != unit->y) {
		if (dside->designer) {
			/* Teleport the unit to its new location. */
			if (can_occupy_cell(unit, x, y)) {
				leave_hex(unit);
				enter_hex(unit, x, y);
			} else {
				nearest_unit(map, h, v, &other);
				if (other != NULL && can_occupy(unit, other)) {
					leave_hex(unit);
					enter_transport(unit, other);
				} else {
					SysBeep(20);
				}
			}
		} else {
			if (unit->act && unit->plan) { /* (should be more sophisticated test?) */
				if (distance(unit->x, unit->y, x, y) == 1) {
					if (unit_at(x, y) != NULL) {
						/* There are units at our desired destination. */
						nearest_unit(map, h, v, &other);
						if (other == NULL) {
							/* We've indicated that we want to stack instead of enter. */
							if (can_occupy_cell(unit, x, y)) {
								prep_move_action(unit, x, y, 0, unit);
							} else {
								SysBeep(20);
							}
						}
						if (other->side == unit->side) {
							/* One of ours, maybe get on it. */
							if (can_occupy(unit, other)) {
								prep_enter_action(unit, other, unit);
							} else if (can_occupy(other, unit)) {
								/* Have other unit do an enter action, then move. */
								/* (not quite right, move should happen after other unit
									is actually inside, in case it fills dest) */
								prep_enter_action(other, unit, other);
								order_moveto(unit, x, y);
							} else if (can_occupy_cell(unit, x, y)) {
								order_moveto(unit, x, y);
							} else {
								SysBeep(20);
							}
							didamove = TRUE;
						} else {
							/* Somebody else's. */
							if (0 /* could capture */) {
								/* do_capture_action(unit, other); */
								didamove = TRUE;
							} else {
								prep_attack_action(unit, other, 100);
								didamove = TRUE;
							}
						}
					} else {
						if (can_occupy_cell(unit, x, y)) {
							prep_move_action(unit, x, y, 0, unit);
						} else {
							SysBeep(20);
						}
					}
				} else {
					DGprintf("Ordering %s to move to %d,%d\n", unit_desig(unit), x, y);
					order_moveto(unit, x, y);
					didamove = TRUE;
				}
			}
		}
	}
}

fire_the_selected_unit(map, unit, h, v)
Map *map;
Unit *unit;
int h, v;
{
	int x, y;
	Unit *other;

	nearest_cell(map, h, v, &x, &y);
	if (x != unit->x || y != unit->y) {
		if (unit->act && unit->plan) { /* (should be more sophisticated test?) */
			if ((other = unit_at(x, y)) != NULL) {
				/* There's a unit to fire at. */
				if (other->side == unit->side) {
					SysBeep(20);
				} else {
					prep_fire_at_action(unit, other, 0);
				}
			} else {
				SysBeep(20);
			}
		}
	}
}

select_next_unit(map)
Map *map;
{
    Unit *thisunit, *nextunit;

	if (map->numselections > 0) {
		thisunit = map->selections[0];
	} else {
		thisunit = NULL;
	}
	nextunit = find_next_unit(dside, thisunit);
	if (nextunit != NULL) {
		unselect_all(map);
		select_unit_on_map(map, nextunit);
		scroll_to_unit(map, nextunit);
		draw_selections(map);
	}
}

select_previous_unit(map)
Map *map;
{
    Unit *thisunit, *prevunit;

	if (map->numselections > 0) {
		thisunit = map->selections[0];
	} else {
		thisunit = NULL;
	}
	prevunit = find_prev_unit(dside, thisunit);
	if (prevunit != NULL) {
		unselect_all(map);
		select_unit_on_map(map, prevunit);
		scroll_to_unit(map, prevunit);
		draw_selections(map);
	}
}

select_next_actor(map)
Map *map;
{
    Unit *thisunit, *nextunit;

	if (map->numselections > 0) {
		thisunit = map->selections[0];
	} else {
		thisunit = NULL;
	}
	nextunit = find_next_actor(dside, thisunit);
	if (nextunit != NULL) {
		unselect_all(map);
		select_unit_on_map(map, nextunit);
		scroll_to_unit(map, nextunit);
		draw_selections(map);
	}
}

select_previous_actor(map)
Map *map;
{
    Unit *thisunit, *prevunit;

	if (map->numselections > 0) {
		thisunit = map->selections[0]; /* should look for an actor */
	} else {
		thisunit = NULL;
	}
	prevunit = find_prev_actor(dside, thisunit);
	if (prevunit != NULL) {
		unselect_all(map);
		select_unit_on_map(map, prevunit);
		scroll_to_unit(map, prevunit);
		draw_selections(map);
	}
}

Unit *
find_next_mover(side, prevunit)
Side *side;
Unit *prevunit;
{
	Unit *unit = NULL;

	if (side != NULL) {
		if (prevunit == NULL) prevunit = side->unithead;
		for (unit = prevunit->next; unit != prevunit; unit = unit->next) {
			if (is_unit(unit) && unit->id > 0
				&& alive(unit)
				&& inside_area(unit->x, unit->y)
				&& (unit->act && unit->act->acp > 0)) {
				return unit;
			}
		} 
	}
	return NULL;
}

select_next_mover(map)
Map *map;
{
    Unit *thisunit, *nextunit;

	if (map->numselections > 0) {
		thisunit = map->selections[0];
	} else {
		thisunit = NULL;
	}
	nextunit = find_next_mover(dside, thisunit);
	if (nextunit != NULL) {
		unselect_all(map);
		select_unit_on_map(map, nextunit);
		scroll_to_unit(map, nextunit);
		draw_selections(map);
	}
}

select_previous_mover(map)
Map *map;
{
	SysBeep(20);
#if 0
    Unit *thisunit, *nextunit;

	if (map->numselections > 0) {
		thisunit = map->selections[0];
	} else {
		thisunit = NULL;
	}
	nextunit = find_next_mover(dside, thisunit);
	if (nextunit != NULL) {
		unselect_all(map);
		select_unit_on_map(map, nextunit);
		scroll_to_unit(map, nextunit);
		draw_selections(map);
	}
#endif
}

/* Scroll the given map over to display the given unit. */

scroll_to_unit(map, unit)
Map *map;
Unit *unit;
{
	if (inside_area(unit->x, unit->y)) {
		if (!in_middle(map, unit->x, unit->y)) {
			map->vcx = unit->x;  map->vcy = unit->y;
			center_on_focus(map);
			force_map_update(map);
		}
	}
}

/* This routine changes a map's viewport and magnification to fit the given rectangle. */

zoom_to_fit(map, x1, y1, x2, y2)
Map *map;
int x1, y1, x2, y2;
{
	int wid, hgt, wanted, power;

	printf("Zooming map to fit in area %d,%d - %d,%d\n", x1, y1, x2, y2);
	/* (still need to do y/2 correction) */
	wid = abs(x2 - x1) + 1;  hgt = abs(y2 - y1) + 1;
	map->vcx = min(x1, x2) + wid / 2;  map->vcy = min(y1, y2) + hgt / 2;
	/* Compute the "ideal" size of a displayed cell. */
	wanted = min(map->pxw / wid, map->pxh / hgt);
	/* Search for the best approximation. */
	for (power = NUMPOWERS-1; power > 0; --power) {
		if (hws[power] < wanted) break;
	}
	if (1 /* changes are visible (might not be, if mag unchanged) */) {
		set_map_power(map, power);
		center_on_focus(map);
		set_map_viewport(map);
		set_map_scrollbars(map);
		force_map_update(map);
	}
}

#ifdef DESIGNERS

int painttype;
int curdepth = 1;

apply_designer_tool(map, h, v, mods)
Map *map;
int h, v, mods;
{
	int x, y, dir, oldt;
	Unit *unit;
	extern short curbgttype, curfid, curelevation, curbrushradius;
	Unit *designer_create_unit();

	nearest_boundary(map, h, v, &x, &y, &dir);
	switch (tooltype) {
		case terraintool:
			/* Dispatch on terrain subtype. */
			switch (t_subtype(curttype)) {
				case cellsubtype:
					oldt = terrain_at(x, y);
					painttype = (curttype == oldt ? curbgttype : curttype);
					paint_cell(dside, x, y, curbrushradius, painttype);
					paint_on_drag(map, h, v, mods);
					break;
				case bordersubtype:
					/* Toggle border on first mouse down. */
					paint_border(dside, x, y, dir, curttype, -1);
					/* Dragging then adds or removes, depending on toggle's result. */
					border_on_drag(map, h, v, mods, border_at(x, y, dir, curttype));
					break;
				case connectionsubtype:
					/* Toggle connection on first mouse down. */
					paint_connection(dside, x, y, dir, curttype, -1);
					/* Dragging then adds or removes, depending on toggle's result. */
					connect_on_drag(map, h, v, mods, connection_at(x, y, dir, curttype));
					break;
				case coatingsubtype:
					paint_coating(dside, x, y, curbrushradius, curttype, curdepth);
					paint_coating_on_drag(map, h, v, mods);
					break;
				default:
					run_warning("Garbage %d subtype %d?\n", curttype, t_subtype(curttype));
					break;
			}
			return;
		case unittool:
			/* tmp hack */
			if (!type_allowed_on_side(curutype, side_n(cursidenumber))) {
				SysBeep(20);
				return;
			}
			if ((unit = designer_create_unit(dside, curutype, cursidenumber, x, y)) != NULL) {
				/* Make the new unit be the current selection. */
				unselect_all(map);
				select_unit_on_map(map, unit);
				draw_selections_at(map, unit->x, unit->y);
			} else {
				/* Beep if the creation failed for some reason. */
				SysBeep(20);
			}
			break;
		case peopletool:
			paint_people(dside, x, y, curbrushradius, cursidenumber);
			break;
		case materialtool:
			break;
		case featuretool:
			paint_feature(dside, x, y, curbrushradius, curfid);
			break;
		case elevationtool:
			paint_elevation(dside, x, y, curbrushradius, curelevation);
			break;
		case temperaturetool:
			break;
		case cloudstool:
			break;
		case windstool:
			break;
		default:
			SysBeep(20);
	}
}

paint_on_drag(map, h0, v0, mods)
Map *map;
int h0, v0, mods;
{
	Point pt0, pt1, newmouse;
	int h1, v1, drawn = FALSE, x, y;
	Rect tmprect;

	SetPt(&pt0, h0, v0);
	SetPt(&pt1, h0, v0);
	SetRect(&tmprect, h0, v0, h0, v0);
	/* (should be a generic subr?) */
	while (WaitMouseUp()) {
		GetMouse(&newmouse);
		if (!EqualPt(pt1, newmouse) /* && PtInRect(newmouse, &(map->window->portRect)) */) {
			pt1 = newmouse;
			nearest_cell(map, pt1.h, pt1.v, &x, &y);
			if (in_area(x, y)) {
				paint_cell(dside, x, y, curbrushradius, painttype);
			}
		}
	}
}

border_on_drag(map, h0, v0, mods, paintmode)
Map *map;
int h0, v0, mods, paintmode;
{
	Point pt0, pt1, newmouse;
	int h1, v1, drawn = FALSE, x, y, dir;
	Rect tmprect;

	SetPt(&pt0, h0, v0);
	SetPt(&pt1, h0, v0);
	SetRect(&tmprect, h0, v0, h0, v0);
	/* (should be a generic subr?) */
	while (WaitMouseUp()) {
		GetMouse(&newmouse);
		if (!EqualPt(pt1, newmouse) /* && PtInRect(newmouse, &(map->window->portRect)) */) {
			pt1 = newmouse;
			nearest_boundary(map, pt1.h, pt1.v, &x, &y, &dir);
			if (inside_area(x, y)) {
				paint_border(dside, x, y, dir, curttype, paintmode);
			}
		}
	}
}

connect_on_drag(map, h0, v0, mods, paintmode)
Map *map;
int h0, v0, mods, paintmode;
{
	Point pt0, pt1, newmouse;
	int h1, v1, drawn = FALSE, x, y, dir;
	Rect tmprect;

	SetPt(&pt0, h0, v0);
	SetPt(&pt1, h0, v0);
	SetRect(&tmprect, h0, v0, h0, v0);
	/* (should be a generic subr?) */
	while (WaitMouseUp()) {
		GetMouse(&newmouse);
		if (!EqualPt(pt1, newmouse) /* && PtInRect(newmouse, &(map->window->portRect)) */) {
			pt1 = newmouse;
			nearest_boundary(map, pt1.h, pt1.v, &x, &y, &dir);
			if (inside_area(x, y)) {
				paint_connection(dside, x, y, dir, curttype, paintmode);
			}
		}
	}
}

paint_coating_on_drag(map, h0, v0, mods)
Map *map;
int h0, v0, mods;
{
	Point pt0, pt1, newmouse;
	int h1, v1, drawn = FALSE, x, y;
	Rect tmprect;

	SetPt(&pt0, h0, v0);
	SetPt(&pt1, h0, v0);
	SetRect(&tmprect, h0, v0, h0, v0);
	/* (should be a generic subr?) */
	while (WaitMouseUp()) {
		GetMouse(&newmouse);
		if (!EqualPt(pt1, newmouse) /* && PtInRect(newmouse, &(map->window->portRect)) */) {
			pt1 = newmouse;
			nearest_cell(map, pt1.h, pt1.v, &x, &y);
			if (inside_area(x, y)) {
				paint_coating(dside, x, y, curbrushradius, curttype, curdepth);
			}
		}
	}
}

#endif /* DESIGNERS */
