/* Copyright (c) 1987-1993  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

#include "conq.h"
#include "cmdline.h"

extern int checkpointinterval;
extern int allbedesigners;

/* The structure containing all the startup-settable options. */

Options options;

/* This is a command-line parser that may be used in implementations that
   get a command line from somewhere. */

#define REQUIRE_ONE_ARG  \
  if (i+1 >= argc) {  \
    fprintf(stderr, "Error: \"%s\" requires another argument\n", argv[i]);  \
    exit(1);  \
  } \
  numused = 2;

/* This says whether the command line arguments actually chose a
   specific game to play (as opposed to setting random flags). */

int gamechosen = FALSE;

/* This is the array of asked-for players. */

char *rawplayers[100];
int numrawplayers = 0;

int numextramodules = 0;
Module *extramodules[100];

char *rawvariants[100];
int numrawvariants = 0;

char *programname = "";

int haderror = FALSE;

/* Set the most basic of defaults on the dynamically-settable options. */

/* (need flags to indicate which options were actually used, so variant
   handling can warn about improper use) */

init_options()
{
    options.worldwidth = DEFAULTWIDTH;
    options.worldheight = DEFAULTHEIGHT;
    options.worldseen = FALSE;
    allbedesigners = FALSE;
    options.alltimeout = 0;
    options.totalgametime = 0;
    options.adddefaultplayer = TRUE;
    options.automatedefaultplayer = FALSE;
}

/* Generic command line parsing.  This is used by several different programs,
   so it just collects info, doesn't process it much.  This is called
   several times, because the validity of some args depends on which
   game modules are loaded. */

parse_command_line(argc, argv, spec)
int argc, spec;
char *argv[];
{
    char ch;
    int i, n, numused;
    Module *module;

    programname = argv[0];

    init_options();

    for (i = 1; i < argc; ++i) {
	if (argv[i] == NULL) {
	    /* Already munched, nothing to do. */
	} else if ((argv[i])[0] == '-') {
	    ch = (argv[i])[1];
	    Dprintf("-%c\n", ch);
	    numused = 1;
	    switch (ch) {
	      case 'A':
		if (spec != 3) continue;
		options.automatedefaultplayer = TRUE;
		break;
	      case 'B':
		if (spec != 1) continue;
		allbedesigners = TRUE;
		break;
	      case 'c':
		REQUIRE_ONE_ARG;
		if (spec != 1) continue;
		checkpointinterval = atoi(argv[i+1]);
		break;
#ifdef DEBUGGING
	      case 'D':
		if (spec != 1) continue;
		Debug = TRUE;
		if (strchr(argv[i]+2, '-')) Debug = FALSE;
		if (strchr(argv[i]+2, 'M')) DebugM = TRUE;
		if (strchr(argv[i]+2, 'G')) DebugG = TRUE;
		break;
#endif /* DEBUGGING */
	      case 'e':
		REQUIRE_ONE_ARG;
		if (spec != 3) continue;
		n = atoi(argv[i+1]);
		while (n-- > 0) rawplayers[numrawplayers++] = "mplayer";
		break;
	      case 'f':
		REQUIRE_ONE_ARG;
		if (spec != 1) continue;
		module = get_game_module(NULL);
		/* Have to hope that module gets named during load. */
		module->filename = copy_string(argv[i+1]);
		if (mainmodule == NULL) {
		    mainmodule = module;
		} else {
		    extramodules[numextramodules++] = module;
		}
		gamechosen = TRUE;
		break;
	      case 'g':
		REQUIRE_ONE_ARG;
		if (spec != 1) continue;
		module = get_game_module(copy_string(argv[i+1]));
		if (mainmodule == NULL) {
		    mainmodule = module;
		} else {
		    extramodules[numextramodules++] = module;
		}
		gamechosen = TRUE;
		break;
	      case 'h':  /* assume this is "-help" */
		if (spec != 1) continue;
		general_usage_info();
		exit(0);
		break;
	      case 'L':
		REQUIRE_ONE_ARG;
		if (spec != 1) continue;
		xconqlib = copy_string(argv[i+1]);
		break;
	      case 'M':
		REQUIRE_ONE_ARG;
		if (spec != 2) continue;
		options.worldwidth = atoi(argv[i+1]);
		options.worldheight = options.worldwidth /*atoi(argv[i+1])*/;
		break;
	      case 'n':
		if (spec != 3) continue;
		options.adddefaultplayer = FALSE;
		break;
	      case 'R':
		REQUIRE_ONE_ARG;
		if (spec != 1) continue;
		init_random(atoi(argv[i+1]));
		break;
	      case 't':
		REQUIRE_ONE_ARG;
		if (spec != 1) continue;
		options.alltimeout = 60 * atoi(argv[i+1]);
		break;
	      case 'T':
		REQUIRE_ONE_ARG;
		if (spec != 1) continue;
		options.totalgametime = 60 * atoi(argv[i+1]);
		break;
	      case 'v':
		if (spec != 2) continue;
		options.worldseen = TRUE;
		break;
	      case 'V':
		if (spec != 2) continue;
		REQUIRE_ONE_ARG;
		rawvariants[numrawvariants++] = argv[i+1];
		break;
	      case 'w':
		if (spec != 1) continue;
		options.warnings = FALSE;
		break;
	      case 'W':
		if (spec != 1) continue;
		/* set world circumference */
		break;
	      default:
		numused = 0;
		/* Anything unrecognized during the last parse is an error. */
		if (spec >= 3) {
		    fprintf(stderr, "Unrecognized option \"%s\"\n", argv[i]);
		    haderror = TRUE;
		}
		break;
	    }
	    if (numused >= 1) argv[i] = NULL;
	    if (numused >= 2) argv[i+1] = NULL;
	    if (numused >= 1) i += (numused - 1);
	} else {
	    if (spec == 3) {
		rawplayers[numrawplayers++] = argv[i];
		argv[i] = NULL;
	    }
	}
    }
    if (haderror) {
	general_usage_info();
	exit(1);
    }
}

/* This routine prints out help info about the args parsed previously. */

general_usage_info()
{
    printf("Usage:\n\t%s [ -options ... ]\n\n", programname);
    printf("General options:\n\n");
    printf("    -B\t\t\tmake every human player a designer\n");
    printf("    -c n\t\tcheckpoint every n turns\n");
    printf("    -f filename\t\trun filename as a game\n");
    printf("    -g gamename\t\tfind gamename in library and run it\n");
    printf("    -h[elp]\t\t\tdisplay this help info\n");
    printf("    -L pathname\t\tset pathname to be library location\n");
    printf("    -t seconds\t\tlimit each player to given time total\n");
    printf("    -T seconds\t\t\tlimit each player to given time per turn\n");
    printf("    -w\t\t\tsuppress warnings\n");
    printf("\nGame variant options:\n\n");
    game_usage_info();
    printf("\nPlayer setup options:\n\n");
    player_usage_info();
}

game_usage_info()
{
    printf("    -M width[xheight][+lat][+lon]\t\t\t\n");
    printf("    -v\t\t\tmake the world be seen already\n");
    printf("    -W diameter\t\t\t\n");
    /* Dump the specific variants for this game here. */
}

player_usage_info()
{
    printf("    name,ai/config@host+strength\tadd player with name, config, host\n");
    printf("    -A[uto]\t\t\tuse AI with default player\n");
    printf("    -e number\t\tadd number computer players to player list\n");
    printf("    -n[odefault]\t\t\tdon't add default player\n");
}

/* Load all the game modules that were asked for on cmd line. */

load_all_modules()
{
    int i;

    if (mainmodule != NULL) {
	load_game_module(mainmodule);
	for (i = 0; i < numextramodules; ++i) {
	    load_game_module(extramodules[i]);
	}
    } else {
	/* If we've got absolutely no files to load, the standard game is
	   the one to go for.  It will direct the remainder of random gen. */
	load_standard_game();
    }
}

/* Set module variants from command line options. */

set_variants_from_options()
{
    int dflt;
    char *varname = "?", *vartypename = NULL;
    char buf[BUFSIZE];
    Obj *vars, *var, *vartype, *vardata;

    for (vars = mainmodule->variants; vars != lispnil; vars = cdr(vars)) {
	var = car(vars);
	if (stringp(car(var))) {
	    varname = c_string(car(var));
	    var = cdr(var);
	}
	vartype = car(var);
	vardata = cdr(var);
	if (symbolp(vartype)) vartypename = c_string(vartype);
	switch (keyword_value(vartypename)) {
	  case K_WORLD_SEEN:
	    dflt = c_number(eval(car(vardata)));
	    if (options.worldseen) dflt = !dflt;
	    set_g_terrain_seen(dflt);
	    break;
	  case K_SEE_ALL:
	    dflt = c_number(eval(car(vardata)));
	    if (0) dflt = !dflt;
	    set_g_see_all(dflt);
	    break;
	  case K_WORLD_SIZE:
	    set_area_shape(options.worldwidth, options.worldheight, TRUE);
	    break;
	  case K_FLAG:
	    break;
	  case K_NUMBER:
	    break;
	  case K_EVAL:
	    break;
	  default:
	    sprintlisp(buf, car(vars));
	    init_warning("Unknown variant %s, ignoring", buf);
	}
    }
    /* These are last-gasp options, perhaps shouldn't be here? */
    if (options.alltimeout > 0) {
	set_g_rt_per_side_max(options.alltimeout);
    }
    if (options.totalgametime > 0) {
	set_g_rt_for_game_max(options.totalgametime);
    }
    if (area.width <= 0) {
	set_area_shape(options.worldwidth, options.worldheight, TRUE);
    }
}

/* Set player characteristics from command-line options. */

set_players_from_options()
{
    int i;
    Player *player;

    if (numrawplayers == 0 || options.adddefaultplayer) {
	player = add_default_player();
	if (options.automatedefaultplayer) {
	    player->aitypename = "mplayer";
	}
	canonicalize_player(player);
    }
    for (i = 0; i < numrawplayers; ++i) {
	player = add_player();
	if (strcmp(rawplayers[i], "mplayer") == 0) { /* not really needed? */
	    player->aitypename = "mplayer";
	} else {
	    parse_player_spec(player, rawplayers[i]);
	    player->displayname = rawplayers[i];
	}
	canonicalize_player(player);
    }
}

/* Parse the syntax "username,ai/config@display". */

parse_player_spec(player, spec)
Player *player;
char *spec;
{
    if (spec != NULL && strcmp(spec, "*") != 0) {
	/* Extract a trailing strength specification. */
	if (iindex('+', spec) >= 0) {
	    player->displayname = copy_string(spec + iindex('@', spec) + 1);
	    player->strength = max(1, atoi(&(spec[iindex('+', spec)+ 1])));
	    spec[iindex('+', spec)] = '\0';
	}
	/* Extract a displayname. */
	if (iindex('@', spec) >= 0) {
	    player->displayname = copy_string(spec + iindex('@', spec) + 1);
	    spec[iindex('@', spec)] = '\0';
	    player->name = spec;
	} else {
	    if (strcmp(spec, "mplayer") == 0) {
		player->aitypename = copy_string(spec);
	    } else {
		player->displayname = copy_string(spec);
	    }
	}
	/* Separate user and configuration name if both given. */
	if (iindex('/', player->name) >= 0) {
	    player->configname =
	      copy_string(player->name + iindex('/', spec) + 1);
	    (player->name)[iindex('/', spec)] = '\0';
	}
    }
    canonicalize_player(player);
}
