/* Copyright (c) 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* This river generator requires elevations. */

#include "conq.h"

long totalrivers;

/* Add rivers by picking a headwater randomly, then running downhill. */

/* Can do as either a border or connection type, depending on which
   terrain type is called "river". */

make_rivers()
{
    int x, y, t;

    totalrivers = 0;
    for_all_hexes(x, y) {
	totalrivers += t_river_chance(terrain_at(x, y));
    }
    if (totalrivers <= 0) return;
    totalrivers /= 10000;
    totalrivers = max(1, totalrivers);
    for_all_terrain_types(t) {
	if (t_is_border(t)
	    && (strncmp("river", t_type_name(t), 5) == 0
		|| extension_value(t_extensions(t), "river", FALSE))
	    && !aux_terrain_defined(t)
	    && elevations_defined()) {
	    make_up_river_borders(t);
	} else if (t_is_connection(t)
		   && (strncmp("river", t_type_name(t), 5) == 0
		       || extension_value(t_extensions(t), "river", FALSE))
		   && !aux_terrain_defined(t)
		   && elevations_defined()) {
	    make_up_river_connections(t);
	}
    }
    /* No idea how to make rivers, give up silently. */
}

int touchedwater;

int lastrx, lastry, lastrvdir;

make_up_river_borders(rivertype)
int rivertype;
{
    int i, j, x, y, dir, elev, len, x1, y1, d1, e1, x2, y2, d2, e2;
    int x0, y0, numrivers = 0, numrivs;
    int low, lowdir;
    int t, watery[MAXTTYPES];

    announce_lengthy_process("Creating rivers (as borders)");
    allocate_area_aux_terrain(rivertype);
    for_all_interior_hexes(x0, y0) {
    	if (y0 % 2 != 0 || x0 % 2 != 0) continue;
	if (probability(t_river_chance(terrain_at(x0, y0)) / 100)) {
	    ++numrivers;
	    touchedwater = FALSE;
	    if (random(5) == 0)
	      announce_progress((100 * numrivers) / totalrivers);
	    x = x0;  y = y0;
	    /* Set the initial river direction. */
	    lowdir = 0;
	    low = elev_at_meet(x, y, lowdir);
	    for_all_directions(dir) {
		if (elev_at_meet(x, y, dir) < low) {
		    lowdir = dir;
		    low = elev_at_meet(x, y, lowdir); 
		}
	    }
	    dir = lowdir;
	    /* We do a right-hand -> downhill rule, so may need to flip around
	       and look at/work with river on opposite side. */
	    if (elev_at_meet(x, y, dir) <
		elev_at_meet(wrap(x+dirx[dir]),y+diry[dir], opposite_dir(dir))) {
		x += dirx[dir];  y += diry[dir];
		dir = opposite_dir(dir);
	    }
	    if (border_is_compatible(x, y, dir, rivertype)) {
		set_river_at(x, y, dir, rivertype);
	    } else {
		continue;
	    }
	    for (j = 0; j < 20; ++j) {  /* why 20?? */
		if (!inside_area(x, y)) break;
		elev = elev_at_meet(x, y, dir);
		/* Compute hex and dir of the two possible ways to flow. */
		x1 = x + dirx[right_dir(dir)];  y1 = y + diry[right_dir(dir)];
		d1 = left_dir(dir);
		e1 = elev_at_meet(x1, y1, d1);
		x2 = x;  y2 = y;
		d2 = right_dir(dir);
		e2 = elev_at_meet(x2, y2, d2);
		if (elev < e1 && elev < e2) {
		    init_warning("river going uphill??");
		    break;
		}
		/* Pick the lower of the two. */
		if (e1 < e2) {
		    x = x1;  y = y1;  dir = d1;
		} else {
		    x = x2;  y = y2;  dir = d2;
		}
		/* Rivers always follow the same paths, so if we see a river
		   here already, we're done. */
		if (border_at(x, y, dir, rivertype)) break;
		if (border_is_compatible(x, y, dir, rivertype)) {
		    set_river_at(x, y, dir, rivertype);
		}
	    }
	}
    }
    /* Cells surrounded by river should maybe be set specially. */
    if (g_river_sink_terrain() != NONTTYPE) {
	for_all_hexes(x, y) {
	    if (inside_world(x, y)) {
		numrivs = 0;
		for_all_directions(dir) {
		    if (border_at(x, y, dir, rivertype)) ++numrivs;
		}
		if (numrivs >= NUMDIRS) {
		    set_terrain_at(x, y, g_river_sink_terrain());
		}
	    }
	}
    }
    /* Fix any bad adjacencies that got through. */
    for_all_terrain_types(t) {
    	watery[t] = watery_terrain(t);
    }
    for_all_hexes(x, y) {
	if (inside_world(x, y)) {
	    if (watery[terrain_at(x, y)]) {
		for_all_directions(dir) {
		    set_border_at(x, y, dir, rivertype, FALSE);
		}
	    }
	}
    }
    finish_lengthy_process();
}

border_is_compatible(x, y, dir, b)
int x, y, dir, b;
{
    return (tt_adj_terr_effect(terrain_at(x, y), b) < 0
	    && tt_adj_terr_effect(terrain_at(wrap(x+dirx[dir]), y+diry[dir]), b) < 0);
}

watery_terrain(t)
int t;
{
    return extension_value(t_extensions(t), "water", FALSE);
}

elev_in_dir(x, y, dir)
int x, y, dir;
{
    return elev_at(wrap(x+dirx[dir]), y+diry[dir]);
}

/* The elevation at the junction of three hexes is the lowest of the three. */

elev_at_meet(x, y, dir)
int x, y, dir;
{
    int elev = elev_at(x, y);

    if (elev_in_dir(x, y, dir) < elev)
      elev = elev_in_dir(x, y, dir);
    if (elev_in_dir(x, y, right_dir(dir)) < elev)
      elev = elev_in_dir(x, y, right_dir(dir));
    return elev;
}

/* should have various conditions */

set_river_at(x, y, dir, t)
int x, y, dir, t;
{
    if (touchedwater) {
    } else if (watery_terrain(terrain_at(x, y))
	       || watery_terrain(terrain_at(wrap(x+dirx[dir]),y+diry[dir]))) {
	touchedwater = TRUE;
    } else {
	set_border_at(x, y, dir, t, TRUE);
    }
}

make_up_river_connections(rivertype)
int rivertype;
{
    int i, x, y, dir, j, len, nx, ny, ndir;
    int rvdir, nrvdir;

    allocate_area_aux_terrain(rivertype);
    announce_lengthy_process("Creating rivers (as connections)");
    for (i = 0; i < totalrivers; ++i) {
	announce_progress((100 * i) / totalrivers);
	/* Pick a headwater */
	/* should check nonadjacency with >= two other rivers */
	random_point(&x, &y);
	/* Set the initial river direction */
	rvdir = random_dir();
	touchedwater = FALSE;
	lastrx = x;  lastry = y;  lastrvdir = rvdir;
	len = random(50)+10;
	for (j = 0; j < len && !touchedwater; ++j) {
	    /* Compute a next segment */
	    /* Rivers prefer to go straight */
	    if (elev_at(x, y) > elev_in_dir(x, y, rvdir) && probability(60)) {
		nrvdir = rvdir;
	    } else if (elev_at(x, y) > elev_in_dir(x, y, left_dir(rvdir))) {
		nrvdir = left_dir(rvdir);
	    } else if (elev_at(x, y) > elev_in_dir(x, y, right_dir(rvdir))) {
		nrvdir = right_dir(rvdir);
	    } else if (elev_at(x, y) > elev_in_dir(x, y, right_dir(right_dir(rvdir)))) {
		nrvdir = right_dir(right_dir(rvdir));
	    } else if (elev_at(x, y) > elev_in_dir(x, y, left_dir(left_dir(rvdir)))) {
		nrvdir = left_dir(left_dir(rvdir));
	    } else if (elev_at(x, y) > elev_in_dir(x, y, rvdir)) {
		/* this in case straight ahead really was the right way */
		nrvdir = rvdir;
	    } else {
		/* can't go anywhere, so make a lake! */
#if 0
		if (terrain_at(x, y) != 0) {
		    set_terrain_at(x, y, 1);
		}
#endif
		/* Done with this river now */
		break;
	    }
	    rvdir = nrvdir;
	    /* Put river in the new direction. */
	    set_river_conn_at(x, y, rvdir, rivertype);
	    /* if we are touching any river other than the last one,
	       should stop */
	    nx = x + dirx[rvdir];  ny = y + diry[rvdir];
	    x = nx;  y = ny;
	}
    }
    /* Clean up a little */
    /* River should not touch anything that is water (or ice!) */
#if 0
    for_all_hexes(x, y) {
	if (terrain_at(x, y) == 0 || terrain_at(x, y) == 1) {
	    for_all_directions(dir) {
		set_border_at(x, y, dir, thebordertype);
	    }
	}
    }
#endif
    finish_lengthy_process();
}

/* should have various conditions */

set_river_conn_at(x, y, dir, rivertype)
int rivertype;
{
    if (touchedwater
	|| terrain_at(x, y) == 0
	|| terrain_at(wrap(x+dirx[dir]),y+diry[dir]) == 0) {
	touchedwater = TRUE;
    } else if (0) {
	touchedwater = TRUE;
    } else {
	set_connection_at(x, y, dir, rivertype, TRUE);
	lastrx = x;  lastry = y;  lastrvdir = dir;
    }
}
