/* Copyright (c) 1987, 1988, 1991  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* All the definitions needed by the machine code. */

/* Determine how many hexes across one area of control is for the machine.  */

#define CAREA_SIZE 10

/* An area of control is a block that is treated as a single */
/* location. All information on the whole area is combined so that the */
/* machine can do some planning.  */

typedef struct a_control_area {
  short enemy_strength;   /* Estimate of enemy unit strength */
  short enemy_seen_recently; /* How many enemy have we seen in last */
			     /* turn. */
  short units_lost;       /* How many units have we lost here. */
			  /* Weighted average of last few turns. */
  short capturers_approaching; /* number of units comming to capture bases */
  short capture_time;     /* When do we expect to have capturers here */
  short safe_area;        /* do we really need to move units here */
} Control_area;
			     

/* Strategy is what a side uses to make decisions. */

typedef struct a_strategy {
    int type;  /* placeholder */
    int trytowin;
    int creationdate;
    short strengths[MAXSIDES][MAXUTYPES];  /* estimated numbers of units */
    short alstrengths[MAXSIDES][MAXUTYPES];  /* numbers in alliances */

    int analyzegame;

    struct a_theater *theaters;

    struct a_goalstate *goalstates; /* what the side is trying to accomplish */

    /* Exploration and search slots. */
    int zonewidth, zoneheight;
    int numzonex, numzoney;     /* dimensions of search zone array */
    int numzones;
    struct a_searchzone *searchzones;
    short *explorertypes;
    short explorersneeded;
    short *terrainguess;

    short cx, cy;               /* "centroid" of all our units */
    short *demand;              /* worth of each utype w.r.t. strategy */

    Control_area *areas;

    int explore_priority;
    int defend_priority;
    int attack_priority;

    struct a_unit *unitlist[MAXUTYPES];   /* lists to help mplay efficiency */
    short unitlistcount[MAXUTYPES];  /* counts of above lists */

} Strategy;

#define mplayer(s) ((Strategy *) (s)->ai)

#define for_all_side_goalstates(s,gs)  \
  for ((gs) = mplayer(s)->goalstates; (gs) != NULL; (gs) = (gs)->next)

extern int delaymove;

typedef struct a_searchzone {
    int size;
    int unexplored;
} SearchZone;

#define search_zone_at(s,x,y)  \
  (&(mplayer(s)->searchzones[((y)/mplayer(s)->zoneheight) * (mplayer(s)->numzonex) + ((x)/mplayer(s)->zonewidth)]))

#define search_zone_n(s,n,zone,x,y)  \
  (zone) = (mplayer(s)->searchzones + n);  \
  (x) = ((n) % mplayer(s)->numzonex) * mplayer(s)->zonewidth;  \
  (y) = ((n) / mplayer(s)->numzonex) * mplayer(s)->zoneheight;

#define for_all_search_zones(s,i)  \
  for (i = 0; i < mplayer(s)->numzones; ++i)
  
/* A Theater is a large hexagonal area that can be planned for all at once. */

typedef struct a_theater {
    char *name;             /* an informative name for this theater */
    short x, y;             /* center of the theater */
    short radius[NUMDIRS];    /* approx size of the theater */
    short importance;       /* 0 = shrug, 100 = critical */
    GoalType purpose;       /* General notion of what this theater is for */
    GoalState *maingoalstate;
    short allied_units;     /* How many units on our side here. */
    short allied_makers;    /* How many of our makers */
    short neutral_makers;    /* How many of neutral makers */
    short makers;           /* Total number of makers */
    short unexplored;       /* Number of unseem hexes in theater. */
    short allied_bases;     /* Total number of our bases, includes towns */
    short border;           /* True if this is a border theater. */
    short nearby;           /* Is this close to a location we have bases. */
    short reinforce;         /* Priority on request for units. */
    struct a_unit *units[MAXUTYPES]; /* Units in theater */
    short requested_units[MAXUTYPES]; /* units we should move to theater. */
    short num_units[MAXUTYPES]; /* units we should move to theater. */
    short enemy_strength;   /* Estimate of enemy unit strength */
    short enemy_seen_recently; /* How many enemy have we seen in last */
    /* turn. */
    short units_lost;       /* How many units have we lost here. */
    /* Weighted average of last few turns. */
    short capturers_approaching; /* number of units comming to capture bases */
    short capture_time;     /* When do we expect to have capturers here */
    short safe_area;        /* do we really need to move units here */
    struct a_theater *next;
} Theater;

#define for_all_theaters(s,th)  \
  for ((th) = mplayer(s)->theaters; (th) != NULL; (th) = (th)->next)

/* An oparea is a part of a theater. */

typedef struct a_oparea {
    struct a_theater *theater;
    short number;           /* oparea number */
    short x, y;             /* center of oparea */
    short allied_units;     /* How many units on our side here. */
    short allied_makers;    /* How many of our makers */
    short neutral_makers;    /* How many of neutral makers */
    short makers;           /* Total number of makers */
    short unexplored;       /* Number of unseem hexes in oparea. */
    short allied_bases;     /* Total number of our bases, includes towns */
    short border;           /* True if this is a border oparea. */
    short nearby;           /* Is this close to a location we have bases. */
    short unit_request;     /* Priority on request for units. */
    struct a_unit *units[MAXUTYPES]; /* Units in oparea */
    short requested_units[MAXUTYPES]; /* units we should move to oparea. */
    short num_units[MAXUTYPES]; /* units we should move to oparea. */
} Oparea;

/* values for unit_request */
#define NO_UNITS 0
#define GUARD_BASE 1
#define PATROL_OPAREA 2
#define EXPLORE_OPAREA 3
#define DEFEND_OPAREA 4
#define GUARD_TOWN 6
#define GUARD_BORDER 2
#define GUARD_BORDER_TOWN 10
#define DEFEND_BASE 20
#define DEFEND_TOWN 50

/* Malloced integer array accessors and modifers. */

#define aref(m,x,y) ((m)[(x)+area.width*(y)])

#define aset(m,x,y,v) ((m)[(x)+area.width*(y)] = (v))

/* Var and fn decls. */

extern int unit_count;
extern int worths_known;

extern int evaluate_hex(), maximize_worth();

/* General collections of numbers used by all machine players. */

extern int *fraction;        /* percentages of terrain types in world */
extern int *unit_hexes;      /* hexes of terrain in world unit can move on */
extern int *bw;              /* how much is a unit worth */
extern int *maxoccupant;     /* total capacity of a transport */
extern int *ave_build_time;  /* Average length of time to build a unit */
extern int *localworth;                /* for evaluation of nearby hexes */
#ifdef REGIONS
extern short **unit_region;  
     /* Regions a unit within which unit can move */
extern int **unit_region_size;
    /* How many hexes in each region. not necessarily contiguous,
       but decent estimate.  Allows lakes to be spotted. */
extern short *num_regions;   /* Number of regions for this unit type */
extern short **units_in_region; /* Number of units currently in region */
extern long munit_regions;            /* Regions current unit can reach */
#endif
extern Oparea *oparea_info;       /* info about opareas, shared for all sides */
extern short opareas_wide;      /* number of opareas in x direction */
extern short opareas_high;      /* number of opareas in y direction */
#define NUMTOPOPAREAS 5
extern Oparea *top_opareas[NUMTOPOPAREAS];       /* Opareas requesting the most units. */

extern int can_move_in_dir[NUMDIRS];  /* directions it make sence to try to move in */

/* Some basic ranges for prioritizing a units tasks. */
#define EXPLORE_VAL 2500
#define FAVORABLE_COMBAT 5000
#define UNFAVORABLE_COMBAT 4000
#define CAPTURE_MAKER 50000
#define CAPTURE_OTHER 20000
#define PATROL_VAL 100
#define HEAD_FOR_GOAL 2300
#define MEET_TRANSPORT 3000

extern long bestworth, bestx, besty;

extern Unit *munit;                    /* Unit being decided about */

extern Side *mside;                    /* Side whose unit is being decided about */
extern int base_building;             /* true if base building is possible. */

extern int worths_known;
extern int route_max_distance;

/* in mutil.c */
extern char *unit_desig();
extern int units_nearby();
extern int survive_to_build_base(), exact_survive_to_build_base();
extern int base_nearby(), any_base_nearby();
extern int occupant_can_capture();
extern int occupant_can_capture_neighbor();
extern int find_closest_unit(), can_produce(), can_move();
extern int out_of_ammo();
extern int fullness(), survival_time();
extern int haven_p(), shop_p(), good_haven_p();
extern Task *find_route(), *find_route_aux(), *make_route_chain();
extern Task *make_route_step();
extern int follow_plan();
extern int unit_strength();
extern void update_opareas();

/* flush this eventually */
#define oparea_index(x,y) (((y) / OPAREA_SIZE) * opareas_wide + (x) / OPAREA_SIZE)

Theater *create_theater();
Theater *theater_at();
