/* Copyright (c) 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact. */

/* Unit plan execution. */

#include "conq.h"

#define building(unit) ((unit)->plan && (unit)->plan->abuilding != NULL)

#define uu_make(x,y) (1)

int printing;

int max_range = 10;  /* dummy value to start to prevent debugging crashes. */

char *plantypenames[] = {

#undef  DEF_PLAN
#define DEF_PLAN(code,NAME) NAME,

#include "plan.def"

    NULL
};

/* flush this eventually */
#define AREA_SIZE 10
extern short opareas_wide;
#define oparea_index(x,y) (((y) / AREA_SIZE) * opareas_wide + (x) / AREA_SIZE)

char *task_desig();
TaskOutcome execute_task();

/* Execute the plan. */

execute_plan(unit, try)
Unit *unit;
int try;
{
    Plan *plan = unit->plan;
    TaskOutcome rslt = TASK_UNKNOWN;
    int immediate = TRUE;
    int u = unit->type, u2;
    Side *us = unit->side;

    DMprintf("%s executing plan %s", unit_desig(unit), plan_desig(plan));
    /* Units that are asleep do nothing. */
    if (plan->asleep || plan->reserve) return;
    /* (should blow out if recursion too deep) */
    if (try > 1) DMprintf(" (try #%d)", try);
    DMprintf("\n");
    if (try > 5) {
	run_warning("%s redoing plan too often, putting it in reserve",
		    unit_desig(unit));
	plan->reserve = TRUE;
    }
    /* Assume that unit actually has a plan, decide what sort it is. */
    switch (plan->type) {
      case UNFORMED_PLAN:
	/* We get one chance to form a plan, putting the unit to sleep
	   if it doesn't work. */
	decide_plan(unit->side, unit);
	if (plan->type != UNFORMED_PLAN) {
	    execute_plan(unit, ++try);
	} else {
	    run_warning("%s could not form a plan, putting it to sleep",
			unit_desig(unit));
	    plan->asleep = TRUE;
	}
	break;
      case NO_PLAN:
	break;
      case FOLLOW_ORDERS:
	/* Brainless units just work from the task queue or ask what to do. */
	if (plan->tasks) {
	    /* (should check that doctrine being followed correctly) */
	    rslt = execute_task(unit, plan->tasks);
	    handle_completed_task(unit, plan, rslt);
	} else if (unit->plan->autotask) {
	    decide_tasks(unit);
	    execute_plan(unit, ++try);
	} else if (unit->side && side_has_ai(unit->side)) {
	    /* Give AI a chance to dream up a plan. */
	    decide_plan(unit->side, unit);
	    execute_plan(unit, ++try);
	} else if (doctrine_allows_wait(unit)) {
	    /* Our task is now to get guidance from the side. */
	    wait_for_orders(unit);
	} else {
	    /* We just are not getting any guidance at all... */
	}
	break;
      case OFFENSIVE_PLAN:
      case DEFENSIVE_PLAN:
	if (0 /* unit->plan->maingoal */) {
	} else if (plan->tasks) {
	    /* (should check that doctrine being followed correctly) */
	    rslt = execute_task(unit, plan->tasks);
	    handle_completed_task(unit, plan, rslt);
	    break;
	} else {
	    if (mobile(u)) {
		int range = max(area.width, area.height);

		if (probability(50)) range = min(range, 10 * u_speed(u));
		/* No special goal, look for something to fight with. */
		if (go_after_victim(unit, range)) {
		    rslt = execute_task(unit, plan->tasks);
		    handle_completed_task(unit, plan, rslt);
		    break;
		}
	    } else if (can_fire(unit)) {
		/* No special goal, look for something to fight with. */
		if (fire_at_opportunity(unit)) {
		    rslt = execute_task(unit, plan->tasks);
		    handle_completed_task(unit, plan, rslt);
		    break;
		}
	    } else {
		for_all_unit_types(u2) {
		    if (mobile(u2) && uu_acp_to_create(u, u2) > 0 && probability(60)) {
			push_build_task(unit, u2, 5);
			break;
		    }
		}
	    }
	}
	/* If nobody to attack, then maybe explore, so fall through. */
	DMprintf("%s found nothing to attack\n", unit_desig(unit));
      case EXPLORATORY_PLAN:
	if (unit->plan->maingoal) {
	    switch (plan->maingoal->type) {
	      case WORLD_KNOWN:
	      case VICINITY_KNOWN:
		if (mobile(u)) {
		    if (probability(90) && explore_reachable_hex(unit)) {
		    } else {
			DMprintf("Explorer %s walking randomly\n",
				 unit_desig(unit));
			random_walk(unit);
		    }
		} else {
			for_all_unit_types(u2) {
				if (mobile(u2) && uu_acp_to_create(u, u2) > 0 && probability(60)) {
					push_build_task(unit, u2, 2);
					break;
				}
			}
		}
		break;
	      default:
		DMprintf("%s goal %s?\n",
			 unit_desig(unit), goal_desig(unit->plan->maingoal));
		break;
	    }
	} else {
	    if (mobile(u)) {
		if (explore_reachable_hex(unit)) {
		} else {
		    DMprintf("%s walking randomly\n", unit_desig(unit));
		    random_walk(unit);
		}
	    } else {
		for_all_unit_types(u2) {
			if (mobile(u2) && uu_acp_to_create(u, u2) > 0 && probability(60)) {
				push_build_task(unit, u2, 2);
				break;
			}
		}
	    }
	}
	if (plan->tasks) {
	    rslt = execute_task(unit, plan->tasks);
	    handle_completed_task(unit, plan, rslt);
	} else {
	    if (probability(10)) {
	    	DMprintf("no tasks, going to sleep (dunno why)\n");
	    	plan->asleep = TRUE;
	    }
	}
	break;
      case RANDOM_PLAN:
	execute_random_plan(unit);
	break;
      default:
	case_panic("plan type", plan->type);
	break;
    }
}

victim_here(x, y)
int x, y;
{
    int u2 = NONUTYPE, uview;
    Unit *unit2;
    Side *side = tmpunit->side, *oside = NULL;

    if (g_see_all()
	|| cover(side, x, y) > 0 /* should be "guaranteed to see" */) {
	if ((unit2 = unit_at(x, y)) != NULL) {
	    u2 = unit2->type;
	    oside = unit2->side;
	} else {
	    return FALSE;
	}
    } else if (terrain_view(side, x, y) != UNSEEN
	       && (uview = unit_view(side, x, y)) != EMPTY) {
	u2 = vtype(uview);
	oside = side_n(vside(uview));
    } else {
	return FALSE;
    }
    return (is_unit_type(u2)
	    && oside != side
	    && could_hit(tmpunit->type, u2)
	    && uu_damage(tmpunit->type, u2) > 0);
}

/* This routine looks for somebody, anybody to attack. */

go_after_victim(unit, range)
Unit *unit;
int range;
{
    int x, y, rslt;

    tmpunit = unit;
    rslt = search_around(unit->x, unit->y, range, victim_here, &x, &y, 1);
    if (rslt) {
	DMprintf("%s found a likely victim at %d,%d, will go after it\n",
		 unit_desig(unit), x, y);
	/* Set up a task to go after the unit found. */
	set_hit_task(unit, x, y);
    }
    return rslt;
}

set_hit_task(unit, x, y)
Unit *unit;
int x, y;
{
    Task *task = create_task(HIT_UNIT_TASK);

    task->args[0] = x;  task->args[1] = y;

    unit->plan->tasks = task;
}

fire_at_opportunity(unit)
Unit *unit;
{
    int x, y, rslt;

    tmpunit = unit;
    rslt = search_around(unit->x, unit->y, u_range(unit->type),
			 victim_here, &x, &y, 1);
    if (rslt) {
	DMprintf("%s found a likely victim at %d,%d, will shoot at it\n",
		 unit_desig(unit), x, y);
	/* Set up a task to shoot at the unit found. */
	set_hit_task(unit, x, y);
    }
    return rslt;
}

/* This is a semi-testing routine that basically picks something for the
   unit to do without worrying about its validity.  The rest of the
   system should function correctly no matter what this thing comes up with. */

execute_random_plan(unit)
Unit *unit;
{
    if (flip_coin()) {
	/* Pick a random task. */
	switch (random((int) NUMTASKTYPES)) {
	  case 0:
	  case 1:
	  default:
	    if (1) {
		int dir = random_dir(), x1, y1;

		point_in_dir(unit->x, unit->y, dir, &x1, &y1);
		order_moveto(unit, x1, y1, 0);
	    } else {
		/* (eventually gen non-adj and up/down moves also) */
	    }
	    break;
	}
    } else {
	/* Pick a random action. */
	Action action;
	char *argtypestr;
	int rslt;

	bzero(&action, sizeof(Action));
	action.type = (ActionType) random((int) NUMACTIONTYPES);
	argtypestr = actiondefns[(int) action.type].argtypes;
	make_plausible_random_args(argtypestr, 0, action.args);
	unit->act->nextaction = action;
	DMprintf("%s randomly tries %s\n",
		 unit_desig(unit), action_desig(&action));
    }
}

/* This attempts to make some vaguely plausible arguments for an action,
   using the types of each arg as a guide.  It also generates *invalid*
   arguments occasionally, which tests error checking in the actions' code. */

/* (should increase prob of certain "useful" combos, such as adj points) */

make_plausible_random_args(argtypestr, i, args)
char *argtypestr;
long i, args[10];
{
    char argch;
    long arg;

    while (i < strlen(argtypestr) && i < 10) {
	argch = argtypestr[i];
	switch (argch) {
	  case 'n':
	    arg = (flip_coin() ? random(100) : (random(20000) - 10000));
	    break;
	  case 'u':
	    arg = random(numutypes + 2) - 1;
	    break;
	  case 'm':
	    arg = random(nummtypes + 2) - 1;
	    break;
	  case 't':
	    arg = random(numttypes + 2) - 1;
	    break;
	  case 'x':
	    arg = random(area.width + 4) - 2;
	    break;
	  case 'y':
	    arg = random(area.height + 4) - 2;
	    break;
	  case 'z':
	    arg = random(10);
	    break;
	  case 'd':
	    arg = random_dir();
	    break;
	  case 'A':
	  case 'U':
	    /* Cast around for a valid unit. */
	    while ((arg = (long) find_unit(random(numunits))) == 0L
		   && probability(98));
	    break;
	  case 'S':
	    arg = 0;
	    break;
	  default:
	    fprintf(stderr, "Garbled action arg type '%c'\n", argch);
	    arg = 0;
	    break;
	}
	args[i++] = arg;
    }
}

/* This routine calculates a basic plan for the given unit. */

decide_plan(side, unit)
Side *side;
Unit *unit;
{
    /* (should be able to make plan object if missing) */
    if (side != NULL) {
	DMprintf("%s deciding on a plan", unit_desig(unit));
	if (side_has_ai(side)) {
	    DMprintf(", using %s AI", side_desig(side));
	    ai_decide_plan(side, unit);
	    DMprintf(" - plan is now %s", plan_desig(unit->plan));
	}
	/* If the AI didn't decide anything, or if this is a human-run
	   unit, then just say the unit will be following orders. */
	/* (doctrine should be able to ask for assortment of plans.) */
	if (unit->plan->type == UNFORMED_PLAN) {
	    if (side_has_display(side))  {
		DMprintf(", but will just be following orders");
		unit->plan->type = FOLLOW_ORDERS;
		unit->plan->tasks = NULL;
	    } else {
		unit->plan->type = FOLLOW_ORDERS;
		unit->plan->tasks = NULL;
		unit->plan->asleep = TRUE;
	    }
	}
	DMprintf("\n");
    } else {
	/* does this ever happen? seems reasonable */
	/* should do a default plan of some sort */
    }
}

doctrine_allows_wait(unit)
Unit *unit;
{
    int everask = units_doctrine(unit, everaskside);

    return (everask);
}

/* Record the unit as waiting for orders about what to do. */

wait_for_orders(unit)
Unit *unit;
{
    if (!unit->plan->waitingfortasks) {
	++(unit->side->numwaiting);
    }
    unit->plan->waitingfortasks = TRUE;
}

/* This runs if an order-following unit has been told to just pick something
   random to do. */

decide_tasks(unit)
Unit *unit;
{
    /* (should pick a variety of actions or tasks randomly) */
    order_moveto(unit, unit->x+random(3)-1, unit->y+random(3)-1, 0);
}

random_walk(unit)
Unit *unit;
{
    /* (should pick a variety of actions or tasks randomly) */
    order_moveto(unit, unit->x+random(3)-1, unit->y+random(3)-1, 0);
}

/* General routine to wake a unit up (and maybe all its cargo). */

wake_unit(unit, wakeocc)
Unit *unit;
int wakeocc;
{
    Unit *occ;

    if (unit->plan) {
	unit->plan->asleep = unit->plan->reserve = FALSE;
    }
    if (wakeocc) {
	for_all_occupants(unit, occ) wake_unit(occ, wakeocc);
    }
}

/* True if unit sees that is is adjacent to an unfriendly. */

adj_enemy(unit)
Unit *unit;
{
#if 0
    int d, x, y;
    viewdata view;

    for_all_directions(d) {
	x = wrap(unit->x + dirx[d]);  y = unit->y + diry[d];
	if (!indep(unit)) {
	    view = side_view(unit->side, x, y);
	    if (view != EMPTY && enemy_side(side_n(vside(view)), unit->side))
	      return TRUE;
	}
    }
#endif
    return FALSE;
}

/* Set up goals for units that need them. */
/* Goals should differ according to unit's role in group... */

decide_goal(unit)
Unit *unit;
{
}

/* Search for a base to move to.  "Extra" is our optimism level, i.e.
   extra distance we might be willing to consider. */

find_base(unit, pred, extra)
Unit *unit;
int (*pred)();
int extra;
{
#if 0
    int range;
    int ox, oy, ux = unit->x, uy = unit->y;
    
    route_max_distance = range = range_left(unit) + extra;
    route_max_distance = min(area.maxdim, range);
    if ((range * range < mside->numunits) ? 
	(search_area(ux, uy, range, pred, &ox, &oy, 1))  :
	(find_closest_unit(ux, uy, range, pred, &ox, &oy))) {
	order_moveto(unit, ox, oy);
/*	if (munit->plan->move_tries < 3)
	  unit->plan->orders.flags |= SHORTESTPATH;
	unit->plan->orders.flags &=
	  ~(ENEMYWAKE|NEUTRALWAKE|SUPPLYWAKE|ATTACKUNIT);
*/
	DMprintf("(found base at %d,%d)", ox, oy);
	return TRUE;
    }
#endif
    return FALSE;
}

/* See if we're in a bad way, either on supply or hits, and get to safety
   if possible.  */

maybe_return_home(unit)
Unit *unit;
{
#if 0
    int u = unit->type;
    
    DMprintf("%s checking return (low %d moves %d); ",
	     unit_desig(unit),
	     low_supplies(unit), moves_till_low_supplies(unit));
    unit->plan->priority = 1000000;
    if (0 /* (low_supplies(unit)
	 || munit->plan->move_tries > 6
	 || (moves_till_low_supplies(unit) < min(3, unit->mp) &&
	     unit->plan->wakeup_reason != WAKEENEMY))
	&& !can_capture_neighbor(unit)
	&& probability(100) */) {
	DMprintf("low supply, ");
	if (building(unit) && survival_time(unit) > 1 /* unit->schedule */) {
	    DMprintf("decided to sit and produce\n");
	    order_sentry(unit, 1);
	    return TRUE;
	}
	if (find_base(unit, good_haven_p, 0)) {
	    DMprintf("found a good base\n");
	    return TRUE;
	} else if (find_base(unit, haven_p, 0)) {
	    DMprintf("found a base\n");
	    return TRUE;
	} else if (unit->transport != NULL) {
	    DMprintf("in a transport, will sit\n");
	    order_sentry(unit, 1);
#if 0
	} else if (survive_to_build_base(unit) && unit->transport == NULL) {
	    DMprintf("going to build a base\n");
	    push_build_task(unit, machine_product(unit));
	    order_sentry(unit, 1 /* unit->schedule */+1);
	    return TRUE;
#endif
	} else if (find_base(unit, haven_p, 1)) {
	    DMprintf("found a base\n");
	    return TRUE;
	} else {
	    DMprintf("but can't do anything about it\n");
	}
    }
#if 0
    if ((cripple(unit) && probability(98))
	|| probability(100 - ((100 * unit->hp) / u_hp(u)))) {
	/* note that crippled units cannot repair themselves */
	DMprintf("badly damaged, ");
	if (unit->transport && could_repair(u, unit->transport->type)) {
	    DMprintf("%s will repair\n", unit_desig(unit->transport));
	    order_sentry(unit, 1);
	    return TRUE;
	} else {
	    if (find_base(unit, shop_p, 0)) {
		DMprintf("found a base\n");
		return TRUE;
	    } else {
		DMprintf("but no place to repair\n");
	    }
	}
    }
#endif
    if (out_of_ammo(unit) >= 0 && probability(80)) {
	DMprintf("should reload, ");
	if (find_base(unit, haven_p, 0)) {
	    DMprintf("found a base\n");
	    return TRUE;
	} else {
#if 0
	    if (/* hack */ 0 && survive_to_build_base(unit) && 
		unit->transport == NULL && probability(90)) {
		DMprintf("going to build something\n");
		push_build_task(unit, machine_product(unit));
		order_sentry(unit, 1 /* unit->schedule */+1);
	    } else {
		DMprintf("but can't\n");
	    }
#endif
	}
    }
#endif
    DMprintf("no need to return\n");
    return FALSE;
}

/* Return the distance that we can go by shortest path before running out */
/* of important supplies.  Will return at least 1, since we can *always* */
/* move one hex to safety.  This is a worst-case routine, too complicated */
/* to worry about units getting refreshed by terrain or whatever. */

range_left(unit)
Unit *unit;
{
#if 0
    int u = unit->type, r, least = 12345; /* bigger than any real value */
    int tmp;
    
    for_all_material_types(r) {
	if (um_tomove(u, r) > 0) {
	    least = min(least, unit->supply[r] / um_tomove(u, r));
	}
	if (um_consume(u, r) > 0) {
	    tmp = (u_speed(u) * unit->supply[r]) / um_consume(u, r);
	    least = min(least, tmp);
	}
    }
    return (least == 12345 ? 1 : least);
#endif
}

/* Make up a chain of tasks that define how to get to the (possibly
   unexplored) destination without running any undue hazards. */

plan_exploration_route(unit, x, y)
Unit *unit;
int x,y;
{
    int maxdist = distance(unit->x, unit->y, x, y);
    Plan *plan = unit->plan;
    Task *taskchain;

    /* Nothing to do? */
    if (maxdist == 0) return;
    /* (should limit maxdist by various heuristics) */

    if ((taskchain = (Task *) find_route(unit, maxdist, x, y)) != NULL) {
	plan->tasks = taskchain;
	DMprintf("%s will explore using %s\n",
		 unit_desig(unit), plan_desig(plan));
    }
}

/* Try to get to the given destination using the router in preference
   to the order_moveto command */

route_to(unit, x, y)
Unit *unit;
int x,y;
{
#if 0
    int maxdist = moves_till_low_supplies(unit);
    Unit *dest_unit = unit_at(x,y);
    Plan *plan;
    Task *taskchain, *task;

    munit = unit;
    if (dest_unit != NULL && isbase(dest_unit))
      maxdist = range_left(unit);
    if (route_max_distance > maxdist)
      maxdist = route_max_distance;
    maxdist = min(area.maxdim, maxdist);
    if ((taskchain = find_route(munit, maxdist, x, y)) != NULL) {
	unit->plan->tasks = taskchain;
	unit->plan->priority = bestworth;
	/* Find the last plan step that we want to use */
	task = unit->plan->tasks;
	while (maxdist-- > 1 && task != NULL)
	  task = task->next;
	if (task != NULL) {
/*	    free_task(task->next);  for now */
	    task->next = NULL;
	}
	DMprintf("%s has tasks going to (%d,%d)\n", unit_desig(unit), x, y);
    } else {
	DMprintf("%s no tasks, moving directly to (%d,%d)\n",
		 unit_desig(unit), x, y);
	order_moveto(unit, x, y);
    }
#endif
}

/* Estimate the goodness and badness of hexes in the immediate vicinity. */

find_worths(range)
int range;
{
#if 0
    int ux = munit->x, uy = munit->y;
    
    mside = munit->side;
    bestworth = -10000;
#ifdef REGIONS
    munit_regions = regions_around(munit->type, ux, uy, FALSE);
#endif
    max_range = range;
    apply_to_area(ux, uy, range, evaluate_hex);
    apply_to_area(ux, uy, range, maximize_worth);
#endif
}

/* This is a heuristic estimation of the value of one unit type hitting */
/* on another.  Should take cost of production into account as well as the */
/* chance and significance of any effect. */

attack_worth(unit, e)
Unit *unit;
int e;
{
    int u = unit->type, worth;

    worth = uu_bhw(u, e);
    if (uu_damage(u, e) >= u_hp(e))
	worth *= 2;
    if (uu_damage(e, u) >= unit->hp)
	worth /= (could_capture(u, e) ? 1 : 4);
    if (could_capture(u, e)) worth *= 4;
    return worth;
}

/* Support functions. */

/* Return true if the given position is threatened by the given unit type. */

threat(side, u, x0, y0)
Side *side;
int u, x0, y0;
{
    int d, x, y, thr = 0;
    Side *side2;
    int view;

    for_all_directions(d) {
	x = wrap(x0 + dirx[d]);  y = y0 + diry[d];
	view = 0 /* side_view(side, x, y) */;
	if (view != UNSEEN && view != EMPTY) {
	    side2 = side_n(vside(view));
	    if (allied_side(side, side2)) {
		if (could_capture(u, vtype(view))) thr += 1000;
		if (uu_bhw(u, vtype(view)) > 0) thr += 100;
	    }
	}
    }
    return thr;
}

/* Heuristics for units to build "bases". */

should_build_base(unit)
Unit *unit;
{
#if 0
    DMprintf(
        "%s should build base survive %d base near %d wakeup %d - ",
		       unit_desig(unit),
		       survive_to_build_base(unit), base_nearby(unit,1),
		       unit->plan->wakeup_reason);
    if ((exact_survive_to_build_base(unit)) && 0 &&
	unit->transport == NULL &&
	!any_base_nearby(unit,1) &&
	probability(90) &&
	unit->plan->wakeup_reason != WAKEENEMY) {
#if 0
	push_build_task(unit, machine_product(unit));
	order_sentry(unit, 1 /* unit->schedule */+1);
	DMprintf("going to build a %d\n", unit->product);
#endif
	return TRUE;
    }
    if ((survive_to_build_base(unit) && 0 &&  probability(20)) &&
	unit->transport == NULL &&
	!any_base_nearby(unit,1) &&
	probability(90) &&
	!neutral_base_nearby(unit, 5) &&
	unit->plan->wakeup_reason != WAKEENEMY) {
#if 0
	push_build_task(unit, machine_product(unit));
	order_sentry(unit, 1 /* unit->schedule */+1);
	DMprintf("going to build a %d\n", unit->product);
#endif
	return TRUE;
    }
    DMprintf("decided not to.\n");
#endif
    return FALSE;
}

get_unit_to_area(unit)
Unit *unit;
{
#if 0
    /* Hack for now */
    unit->plan->type = APPROACH;
    unit->plan->x = area_info[unit->plan->area].x;
    unit->plan->y = area_info[unit->plan->area].y;
    search_for_best_move(unit);
#endif
}

#define UNITPRODUCT 0

/* Decide whether a change of product is desirable. */

change_machine_product(unit)
Unit *unit;
{
#if 0
    int u = unit->type;

    if (building(unit) && probability(3)) {
	if ((unit->built > 2) ||
	    ((uu_make(u, UNITPRODUCT) * unit->built) > 20)) {
	}
	    return TRUE;
    } else {
    }
    return FALSE;
#endif
}

/* Machine algorithm for deciding what a unit should build. This routine */
/* must return the type of unit decided upon.  Variety of production is */
/* important, as is favoring types which can leave the builder other than */
/* on a transport.  Capturers of valuable units are also highly preferable. */

int
machine_product(unit)
Unit *unit;
{
    int u = unit->type, u2, u3,
            type, t, d, x, y, x1, y1, d1, value, bestvalue, besttype, tmp;
    int adjterr[MAXUTYPES];
    int units_close[MAXUTYPES];
    int demand[MAXUTYPES], total_demand, ucount[MAXUTYPES];
    int Values[MAXUTYPES];
    int others, total_units;
    Unit *unit2;

    int enemy_strength;
    int units_lost;
    int defend_priority;
    int explore_priority;
    int attack_priority;

    if (!unit->side || !side_has_ai(unit->side)) return NONUTYPE;
#if 0
    enemy_strength = side_strategy(unit->side)->areas[oparea_index(unit->x, unit->y)].enemy_strength;
    units_lost = side_strategy(unit->side)->areas[oparea_index(unit->x, unit->y)].units_lost;
    defend_priority = side_strategy(unit->side)->defend_priority;
    explore_priority = side_strategy(unit->side)->explore_priority;
    attack_priority = side_strategy(unit->side)->attack_priority;

    mside = unit->side;
    besttype = 0 /* period_firstptype() */;
    /* make sure we find a product if the unit can build. */
    if (!could_make(u, besttype)) {
      for_all_unit_types(u2)
	if (could_make(u, u2))
	  besttype = u2;
    }
    bestvalue = 0;
#ifdef REGIONS
    for_all_unit_types(u2) {
      adjterr[u2] = 0;
      units_close[u2] = 0;
      for_all_directions(d) {
	x = wrap(unit->x + dirx[d]);  y = unit->y + diry[d];
	if ((t = aref(unit_region[u2], x, y)) >= 0) {
	  tmp = TRUE;
	  for (d1 = 0; d1 < d; d1++) {
	    x1 = wrap(unit->x + dirx[d1]);  y1 = unit->y + diry[d1];
	    if (t == aref(unit_region[u2], x1, y1))
	      tmp = FALSE;
	  }
	  if (tmp) {
	    adjterr[u2] += unit_region_size[u2][t];
	    units_close[u2] += units_in_region[u2][t];
	  }
	}
      }
    }
#else
    for_all_unit_types(u2) {
      adjterr[u2] = 0;
      units_close[u2] = 0;
      for_all_directions(d) {
	x = wrap(unit->x + dirx[d]);  y = unit->y + diry[d];
	if (could_move(u2, terrain_at(x, y))) {
	  adjterr[u2] += 1;
	}
      }
      units_close[u2] = units_nearby(x, y, 7, munit->type);
    }
#endif
    if (0) {
      for_all_unit_types(u2) {
	tmp = u_is_base(u2);  /* really should be checking that can leave
				      either by moving or on transport. Bases
				      can always be made. */
	if (could_make(u, u2)) {
	  value = side_strategy(mside)->demand[u2];
	  if (mobile(u2)) {
#ifdef REGIONS
	    value += adjterr[u2] * num_regions[u2] * 
	             600 / (unit_hexes[u2]+1);
#endif
	    if (adjterr[u2] > 0) tmp = TRUE;
	  } else tmp = TRUE;
	  value *= isqrt(u_speed(u2));
	  for_all_unit_types(u3)
	    if (uu_bcw(u2, u3) > 100)
	      value += value / 8;
	  if (u_is_transport(u2) && !u_is_carrier(u2))
	    value += value/4;
	  if (value > 0 && u_is_carrier(u2) && base_building)
	    value /= 2;
	  if (u_is_base(u2)) value += value;
	  if (value > 0)
	    value /= (mside->units[u2] + mside->unitsbuilding[u2] + 1);
	  others = units_nearby(unit->x, unit->y, min(u_speed(u2) * 5, 14), u2);
	  if (others > 0 && value > 0) value /= others;
	  
	  /* might want to adjust by number of existing units? */
	  value = (value * (100 - build_time(unit, u2))) / 100;
	  if (build_time(unit, u2) < 10)
	    value += (value * (10 - build_time(unit, u2))) / 10;
/*	  if (u_selfdestruct(u2) && value > 0) value /= 10;  */
#if 0
	  if (u_research(u2) > 0) {
	    if (mside->unitsbuilding[u2] > 0) {
	      if (value > 0)
		value = value / 100 - 1000;
	      else value -= 1000;
	    } else for_all_unit_types(u3) {
	      if (uu_tech_crossover(u2, u3) && mside->counts[u3] <= 1) {
		if (value > 0)
		  value = value / 100 - 1000;
		else value -= 1000;
	      }
	    }
	}
#endif
	  if (tmp && value > bestvalue) {
	    besttype = u2;
	    bestvalue = value;
	  }
	}
      }
    } else {
      total_demand = 1; /* make sure we don't divide by zero */
      for_all_unit_types(u2) 
	if (could_make(u, u2)) {
	  demand[u2] = u_attack_worth(u2) * attack_priority +
	    u_defense_worth(u2) *
	      (defend_priority + 5 * (enemy_strength + units_lost)) +
	      u_explore_worth(u2) * explore_priority;
	  /* if we are under attack, build quick units. */
	  if (uu_make(u, u2) < 10 && enemy_strength > 3)
	    demand[u2] += demand[u2] / 2;
	  if (uu_make(u, u2) < 5 && enemy_strength > 3)
	    demand[u2] += demand[u2] / 2;
	  total_demand += demand[u2];
	  ucount[u2] = 0;
      }
      total_units = 100;  /* actually hundreths of units */
      for_all_side_units(mside, unit2) {
	if (could_make(u, unit2->type))
	  ucount[unit2->type] += 100;
#if 0
	if (unit != unit2 && building(unit2) && !cripple(unit2)) {
	  ucount[UNITPRODUCT] +=
	    max(30, ((100 + uu_make(unit2->type, UNITPRODUCT) - 
		      unit2->schedule) * 10) /
		max(1,uu_make(unit2->type, UNITPRODUCT)));
	}
#endif
      }
      for_all_unit_types(u2)
	 if could_make(u, u2) {
	   ucount[u2] =  
	     region_portion(ucount[u2], u2, units_close, adjterr);
	   total_units += ucount[u2];
	 }
      for_all_unit_types(u2) {
	tmp = u_is_base(u2);  /* really should be checking that can leave
				      either by moving or on transport. Bases
				      can always be made. */
	if (could_make(u, u2)) {
	  value = demand[u2] - ((3 + ucount[u2]) * total_demand) / total_units;
/*	  if (UNITPRODUCT == u2) value += max(1, value / 4);  */
/*	  others = units_nearby(unit->x, unit->y,
				min(u_speed(u2) * 5, 14), u2); */
/*	  if (others > 0) value /= isqrt(others); */
	  
	  /* might want to adjust by number of existing units? */
	  Values[u2] = value;
#if 0
	  if (u_research(u2) > 0) {
	    if (mside->unitsbuilding[u2] > 0) {
	      if (value > 0)
		value = value / 100 - 1000;
	      else value -= 1000;
	    } else for_all_unit_types(u3) {
	      if (uu_tech_crossover(u2, u3) && mside->counts[u3] <= 1) {
		if (value > 0)
		  value = value / 100 - 1000;
		else value -= 1000;
	      }
	    }
	  }
#endif
	  if (value > bestvalue && adjterr[u2] >
#ifdef REGIONS
	      u_min_region_size(u2)
#else
	    0
#endif
	      )
	    {
	    besttype = u2;
	    bestvalue = value;
	  }
	}
      }
      type= besttype;
/*      notify(side_n(0),"%d %d | %d %d %d %d %d %d",
	     type, total_units, Values[0], adjterr[0], adjterr[1], 
	     adjterr[2], adjterr[3], adjterr[4]); 
*/
/*      printf("\n%d: unit %s was building %s, now %s value %d\n", 
	     curturn, unit_desig(unit),
	     (UNITPRODUCT == NONUTYPE ? "no" : u_type_name(UNITPRODUCT)),
	     (type == NONUTYPE ? "no" : u_type_name(type)),
	     bestvalue); */
    }
    type = besttype;
#endif
    /* safety check */
    if (!could_make(unit->type, type)) type = NONUTYPE;
    DMprintf("%d: %s will now build %s units\n",
	     curturn, unit_desig(unit),
	     (type == NONUTYPE ? "no" : u_type_name(type)));
    return type;
}

/* Take into account other units already in this region.  This will
   tend to increase the unit count of regions which already have more
   units than the portion of the land they account for. */

int
region_portion(n, u, units_close, adjterr)
int n, u, units_close[MAXUTYPES], adjterr[MAXUTYPES];
{
    int result;
#if 0
#ifdef REGIONS
    if (unit_hexes[u] != 0) 
      result = n - (n * adjterr[u]) / (3 * unit_hexes[u]) +
	(n * units_close[u]) / 
	  (3 * max(side_strategy(mside)->unitlistcount[u], 1));
    else 
#endif                
      result = n + (n * units_close[u]) /
	(3 * max(side_strategy(mside)->unitlistcount[u], 1));
    return result;
#endif
}

handle_completed_task(unit, plan, rslt)
Unit *unit;
Plan *plan;
TaskOutcome rslt;
{
    Task *task = plan->tasks;

    DMprintf("%s did task %s: ", unit_desig(unit), task_desig(task));
    /* Now look at what happened with the task. */
    switch (rslt) {
      case TASK_UNKNOWN:
	DMprintf("unknown outcome???\n");
	break;
      case TASK_FAILED:
	DMprintf("failed - try %d - ", task->retrynum);
	if (task->retrynum > 3) {
	    DMprintf("removing it\n");
	    pop_task(plan);
	    /* We might be buzzing, so maybe go into reserve for next turn */
	    if (probability(10)) plan->reserve = TRUE;
	} else {
	    DMprintf("will retry\n");
	}
	break;
      case TASK_INCOMPLETE:
	/* Leave the task alone. */
	DMprintf("is incomplete\n");
	break;
      case TASK_PREPPED_ACTION:
	/* Mention the action prepared to execute. */
	DMprintf("prepped action %s\n", action_desig(unit->act->nextaction));
	break;
      case TASK_COMPLETE:
	DMprintf("complete after %d execs, removing it\n", task->execnum);
	pop_task(plan);
	break;
      default:
	break;
    }
}

pop_task(plan)
Plan *plan;
{
    if (plan->tasks) {
	plan->tasks = plan->tasks->next;
    }
}

react_to_enemies(unit)
Unit *unit;
{
    return (unit->plan ? TRUE : FALSE);
}

/* Unit attempts to follow its leader around, but not too closely. */

/* (should find a path if too far away) */

follow_leader(unit, task)
Unit *unit;
Task *task;
{
    Unit *leader = unit->plan->boss; /* should stash id instead? */

    if (leader == NULL) {   /* find unit won't return dead leader */
/*	wake_unit(unit, FALSE, WAKELEADERDEAD, (Unit *) NULL);  */
	return FALSE;
    } else if (abs(unit->x - leader->x) < 2 && abs(unit->y - leader->y) < 2) {
	unit->act->mp = 0; /* a hack */
	return TRUE;
    } else {
	return do_move_action(unit, leader->x, leader->y, 0, unit);
    }
}


/* Patrol just does move_to, but cycling waypoints around when the first */
/* one has been reached. */

move_patrol(unit)
Unit *unit;
{
#if 0
    int tx, ty;

    if (unit->plan->orders.rept-- > 0) {
	if (unit->x == unit->plan->orders.p.pt[0].x &&
	    unit->y == unit->plan->orders.p.pt[0].y) {
	    tx = unit->plan->orders.p.pt[0].x;
	    ty = unit->plan->orders.p.pt[0].y;
	    unit->plan->orders.p.pt[0].x = unit->plan->orders.p.pt[1].x;
	    unit->plan->orders.p.pt[0].y = unit->plan->orders.p.pt[1].y;
	    unit->plan->orders.p.pt[1].x = tx;
	    unit->plan->orders.p.pt[1].y = ty;
	}
	return move_to(unit, unit->plan->orders.p.pt[0].x, unit->plan->orders.p.pt[0].y,
		       (unit->plan->orders.flags & SHORTESTPATH));
    }
#endif
    return TRUE;
}

/* Check for units waiting to embark */

check_for_embarkies(unit, top_unit)
Unit *unit, *top_unit;
{
    Unit *occ, *next;
    int timeleft;

    next = top_unit->occupant;
    while (next != NULL) {
	occ = next;
	next = occ->nexthere;
#if 0
	if (occ->plan->orders.type == EMBARK && can_carry(unit, occ)) {
	    timeleft = occ->plan->orders.rept;
	    leave_hex(occ);
	    enter_transport(occ, unit);
/*	      order_sentry(occ, min(timeleft, 5));  */
	}
#endif
    }
}

#if 0
    if (!indep(unit)  /* why this restriction? */
	&& type != NONUTYPE
	&& period_setproduct()
	&& could_make(unit->type, type)
	&& type != unit->product) {
	/* Set this as our new product. */
	unit->product = type;
	unit->next_product = type;
	unit->built = 0;
	/* Update the relevant displays. */
	if (unit->product != NONUTYPE) {
	    unit->side->unitsbuilding[unit->product]--;
/*	    update_unit_type_lists(unit->side, unit->product);  */
	}
	if (type != NONUTYPE) {
	    unit->side->unitsbuilding[type]++;
/*	    update_unit_type_lists(unit->side, type);  */
	}
    }
#endif

/* Basic routine to compute how long a unit will take to build something. */

build_time(unit, prod)
Unit *unit;
int prod;
{
    int schedule = uu_make(unit->type, prod);
    int u, research_delay;

    /* Add penalty (or unpenalty!) for first unit of a type. */
    /* is "counts" a reliable way to test? */
    if (unit->side->counts[prod] <= 1) {
/*	research_delay = ((schedule * u_research(prod)) / 100);  */
	for_all_unit_types(u) {
	    if (unit->side->counts[u] > 1) {
		research_delay -=
		  (uu_make(unit->type, u) * 
		   uu_tech_crossover(prod, u)) / 100;
	    }
	    if (research_delay > 0) {
		schedule += research_delay;
	    }
	}
    }
    return schedule;
}

reset_task_queue(plan)
Plan *plan;
{
    if (!plan) return;
    if (plan->tasks)
      run_warning("Blasting tasks %s etc from plan",
		  task_desig(plan->tasks));
    plan->tasks = NULL;
}

void 
free_plan(plan)
Plan *plan;
{
    /* don't do anything yet */
}

/* Describe a plan succinctly. */

char *planbuf = NULL;

char *
plan_desig(plan)
Plan *plan;
{
    Task *task;
    int extra = 0;

    if (planbuf == NULL) planbuf = (char *) xmalloc(1000 * sizeof(char));
    if (plan == NULL) {
	sprintf(planbuf, "no plan");
    } else if (plan->type == UNFORMED_PLAN) {
	sprintf(planbuf, "unformed plan");
    } else {
	if (plan->tasks) {
	    sprintf(tmpbuf, "");
	    for (task = plan->tasks; task != NULL; task = task->next) {
		if (strlen(tmpbuf) < 100) {
		    strcat(tmpbuf, " ");  strcat(tmpbuf, task_desig(task));
		} else {
		    ++extra;
		}
	    }
	    if (extra > 0) {
		sprintf(tmpbuf+strlen(tmpbuf), " ... %d more ...", extra);
	    }
	} else {
	    sprintf(tmpbuf, "no tasks");
	}
	/* Should be able to say more about the plan */
	sprintf(planbuf, "type %s %s",
		plantypenames[plan->type], goal_desig(plan->maingoal));
	if (plan->asleep) strcat(planbuf, " [asleep]");
	if (plan->reserve) strcat(planbuf, " [reserve]");
	if (plan->reserve) strcat(planbuf, " [waiting for tasks]");
	strcat(planbuf, tmpbuf);
    }
    return planbuf;
}

/* True if unit is in immediate danger of being captured. */
/* Needs check on capturer transport being seen. */

might_be_captured(unit)
Unit *unit;
{
    int d, x, y;
    Unit *unit2;

    for_all_directions(d) {
	x = wrap(unit->x + dirx[d]);  y = unit->y + diry[d];
	if (((unit2 = unit_at(x, y)) != NULL) &&
	    (enemy_side(unit->side, unit2->side)) &&
	    (could_capture(unit2->type, unit->type))) return TRUE;
    }
    return FALSE;
}
