/* $XConsortium: Map.c,v 1.4 91/02/17 16:18:42 converse Exp $ */

/* Copyright	Massachusetts Institute of Technology	1987, 1988
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * M.I.T. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL M.I.T.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include "conq.h"
#include "xtconq.h"

#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include "xtmapP.h"

/* static */ XtResource resources[] = {
#define offset(field) XtOffsetOf(MapRec, map.field)
    /* {name, class, type, size, offset, default_type, default_addr}, */
    { XtNside, XtCside, XtRString, sizeof(String),
	offset(side), XtRString, (XtPointer) NULL },
    { XtNvcx, XtCvcx, XtRInt, sizeof(int),
	offset(vcx), XtRString, (XtPointer) "10" },
    { XtNvcy, XtCvcy, XtRInt, sizeof(int),
	offset(vcy), XtRString, (XtPointer) "10" },
    { XtNpower, XtCpower, XtRInt, sizeof(int),
	offset(power), XtRString, (XtPointer) "4" },
    { XtNshowTerrain, XtCshowTerrain, XtRBoolean, sizeof(Boolean),
	offset(drawterrain), XtRString, (XtPointer) "on" },
    { XtNshowPatterns, XtCshowPatterns, XtRBoolean, sizeof(Boolean),
	offset(resource), XtRString, (XtPointer) "off" },
    { XtNshowGrid, XtCshowGrid, XtRBoolean, sizeof(Boolean),
	offset(drawhexedges), XtRString, (XtPointer) "on" },
    { XtNshowUnits, XtCshowUnits, XtRBoolean, sizeof(Boolean),
	offset(drawunits), XtRString, (XtPointer) "on" },
    { XtNshowNames, XtCshowNames, XtRBoolean, sizeof(Boolean),
	offset(drawnames), XtRString, (XtPointer) "on" },
    { XtNforeground, XtCforeground, XtRPixel, sizeof(Pixel),
	offset(fgcolor), XtRString, XtDefaultForeground },
    { "?", "??", XtRString, sizeof(String),
	offset(resource), XtRString, (XtPointer) "default" }
#undef offset
};

/* static */ void MapAction(/* Widget, XEvent*, String*, Cardinal* */);

/* static */ void initialize_map();
/* static */ void realize_map();
/* static */ void redisplay_map();
/* static */ Boolean set_values_map();

/* static */ XtActionsRec actions[] =
{
  /* {name, procedure}, */
    {"map",	MapAction},
};

/* static */ char translations[] =
"<Key>:		map()	\n\
";

MapClassRec mapClassRec = {
  { /* core fields */
    /* superclass		*/	(WidgetClass) &widgetClassRec,
    /* class_name		*/	"Map",
    /* widget_size		*/	sizeof(MapRec),
    /* class_initialize		*/	NULL,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	initialize_map,
    /* initialize_hook		*/	NULL,
    /* realize			*/      realize_map,
    /* actions			*/	actions,
    /* num_actions		*/	XtNumber(actions),
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	NULL,
    /* resize			*/	NULL,
    /* expose			*/	redisplay_map,
    /* set_values		*/	set_values_map,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	XtInheritSetValuesAlmost,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	translations,
    /* query_geometry		*/	XtInheritQueryGeometry,
    /* display_accelerator	*/	XtInheritDisplayAccelerator,
    /* extension		*/	NULL
  },
  { /* map fields */
    /* empty			*/	0
  }
};

WidgetClass mapWidgetClass = (WidgetClass)&mapClassRec;

/* static */ void
MapAction(widget, event, params, num_params)
Widget widget;
XEvent *event;
String *params;
Cardinal *num_params;
{
    XawMapCallDataRec call_data;

    call_data.event = event;
    call_data.params = params;
    call_data.num_params = *num_params;

    XtCallCallbacks( widget, XtNcallback, (XtPointer)&call_data );
}

/* static */ void
initialize_map(request, new)
Widget request, new;
{
    int r;
    MapWidget mw = (MapWidget) new;
    Side *side = side_from_display(XtDisplay(new));

    mw->map.side = side;

    mw->map.gc = (GC) NULL;

    mw->map.isworld = FALSE;

    mw->map.lastvcx = mw->map.lastvcy = -1;

    mw->map.drawpeople = TRUE;
    mw->map.drawelevations = FALSE;
    /* should dynalloc this array */
    for_all_material_types(r) {
	mw->map.drawresources[r] = FALSE;
    }
    mw->map.drawtemp = FALSE;
    mw->map.drawweather = FALSE;

    mw->map.vcx = mw->map.vcy = 10;

    set_map_power(mw);
    set_map_viewport(mw);

    set_colors(side, mw);
}

/* static */ void
realize_map(w, valuemaskp, attr)
Widget w;
XtValueMask *valuemaskp;
XSetWindowAttributes *attr;
{
    XGCValues v;
    MapWidget mw = (MapWidget) w;

    (*mapWidgetClass->core_class.superclass->core_class.realize)
      (w, valuemaskp, attr);
}

/* static */ Boolean
set_values_map(oldw, reqw, w, args, numargs)
Widget oldw, reqw, w;
ArgList args;
Cardinal *numargs;
{
    MapWidget oldmw = (MapWidget) oldw;
    MapWidget reqmw = (MapWidget) reqw;
    MapWidget mw = (MapWidget) w;
    Boolean redisplay = FALSE;

#define NEQ(field) (oldmw->field != mw->field)

    if (NEQ(map.vcx) || NEQ(map.vcy)) {
	if (inside_area(mw->map.vcx, mw->map.vcy)) {
	    /* check validity just in case? */
	    set_map_viewport(mw);
	    redisplay = TRUE;
	} else {
	    mw->map.vcx = oldmw->map.vcx;  mw->map.vcy = oldmw->map.vcy;
	}
    }
    if (NEQ(map.power)) {
	if (between(0, mw->map.power, NUMPOWERS-1)) {
	    set_map_power(mw);
	    set_map_viewport(mw);
	    redisplay = TRUE;
	} else {
	    /* restore the old value, don't redisplay */
	    mw->map.power = oldmw->map.power;
	}
    }
    if (NEQ(map.drawterrain)
	|| NEQ(map.drawhexedges)
	|| NEQ(map.drawhexpats)
	|| NEQ(map.drawunits)
	|| NEQ(map.drawnames)
	|| NEQ(map.drawpeople)
	|| NEQ(map.drawweather)) {
	redisplay = TRUE;
    }
    return (redisplay);
#undef NEQ
}

/* All the kinds of events we're normally interested in. */

int everything = KeyPressMask|ButtonPressMask|ButtonReleaseMask|ExposureMask;

/* various bitmap definitions. */

#include "../x11/snowy"

#include "../x11/lookglass"
#include "../x11/lookmask"
#include "../x11/movecursor"
#include "../x11/movemask"
#include "../x11/shootcursor"
#include "../x11/shootmask"
/* need masks for these too */
#include "../x11/hexcursor"
#include "../x11/bordcursor"
#include "../x11/conncursor"
#include "../x11/addunits"

#include "../x11/boxcurs"
#include "../x11/boxmask"

#include "../x11/bomb1"
#include "../x11/bomb2"
#include "../x11/bomb3"
#include "../x11/bomb4"

#include "../x11/miss"
#include "../x11/hit"
#include "../x11/kill"

#include "../x11/hex8"
#include "../x11/hex8b"

#include "../x11/hex16"
#include "../x11/hex16b"

#include "../x11/closer"
#include "../x11/farther"
#include "../x11/terr"
#include "../x11/elev"
#include "../x11/therm"
#include "../x11/units"
#include "../x11/names"
#include "../x11/people"
#include "../x11/resource"

/* The following magical arrays set all the sizes at each mag. */

/* This is the basic cell size. */

int mags[] = { 1, 2, 4, 8, 16, 32, 64, 128 };

/* These give the total dimensions of each hex cell, plus the vertical
   distance center-to-center.  This is all carefully calculated to make
   the cells overlap perfectly at each different magnification, assuming
   that the icons have the right shape and size. */

int hws[] = { 1, 2, 4, 10, 20, 40, 80, 160 };
int hhs[] = { 1, 2, 5, 11, 22, 44, 88, 176 };
int hcs[] = { 1, 2, 4,  9, 17, 34, 68, 136 };

/* The sizes of the unit subcells. */

int uhs[] = { 1, 2, 2,  8, 16, 32, 64, 128 };
int uws[] = { 1, 2, 2,  8, 16, 32, 64, 128 };

/* Widths of borders (0 implies don't draw at all). */

/* (Sep array for connections?) */

int bwid[] = { 0, 0, 0, 1, 3, 5, 7, 9 };

/* Coordinates of the hex borders. */
/* Note that array has extra entry so don't need to wrap index. */

int bsx[NUMPOWERS][7] = {
    { 0 },
    { 0 },
    { 0 },
    {  5,  10,  10,  5,  0,  0,  5 },
    { 10,  20,  20, 10,  0,  0, 10 },
    { 20,  40,  40, 20,  0,  0, 20 },
    { 40,  80,  80, 40,  0,  0, 40 },
    { 80, 160, 160, 80,  0,  0, 80 }
};
int bsy[NUMPOWERS][7] = {
    { 0 },
    { 0 },
    { 0 },
    {  0,  2,  9, 11,  9,  2,  0 },
    {  0,  4, 18, 22, 18,  4,  0 },
    {  0,  8, 36, 44, 36,  8,  0 },
    {  0, 16, 72, 88, 72, 16,  0 },
    {  0, 32, 144, 176, 144, 32,  0 }
};

/* Coords of middles of each hex border (half a connection, basically). */
 
int lsx[NUMPOWERS][6] = {
    { 0 },
    { 0 },
    { 0 },
    {  2,  5,   2,  -2,  -5,  -2 },
    {  5, 10,   5,  -5, -10,  -5 },
    { 10, 20,  10, -10, -20, -10 },
    { 20, 40,  20, -20, -40, -20 },
    { 40, 80,  40, -40, -80, -40 }
};
int lsy[NUMPOWERS][6] = {
    { 0 },
    { 0 },
    { 0 },
    {  -4,  0,   4,   4,   0,  -4 },
    {  -9,  0,   9,   9,   0,  -9 },
    { -18,  0,  18,  18,   0, -18 },
    { -36,  0,  36,  36,   0, -36 },
    { -72,  0,  72,  72,   0, -72 }
};

/* Do the grody work of drawing very large polygons accurately. */

/* (Worthwhile to precompute all the point arrays?) */

draw_hex_polygon(dpy, win, gc, sx, sy, power, doedges)
Display *dpy;
Window win;
GC gc;
int sx, sy, power, doedges;
{
    XPoint points[6];
    int hw = hws[power], hh = hhs[power], delt = (hhs[power] - hcs[power]);
    int ew = (doedges ? 1 : 0);
	    
    points[0].x = sx + hw / 2;        points[0].y = sy;
    points[1].x = hw / 2 - ew;        points[1].y = delt - ew;
    points[2].x = 0;                  points[2].y = hh - 2 * delt - ew;
    points[3].x = 0 - (hw / 2 - ew);  points[3].y = delt - ew;
    points[4].x = 0 - (hw / 2 - ew);  points[4].y = 0 - (delt - ew);
    points[5].x = 0;                  points[5].y = 0 - (hh - 2 * delt - ew);
    XFillPolygon(dpy, win, gc, points, 6, Convex, CoordModePrevious);
    XFlush(dpy);
}

/* Do misc setup thingies. */

init_misc(side)
Side *side;
{
#if 0
    int u;
    unsigned int w, h;
    unsigned long mask;
    XGCValues values;
    GC gc;
    char *name;
    int junk1, junk2;

    side->ui->margin = margin;
    /* get the name from the display unless the game is being restored from a 
       save file. (?) */
    if (!midturnrestore &&
	(name = XGetDefault(sdd(side), PROGRAMNAME, "SideName")) != NULL)
      side->name = name;
    mask = GCForeground | GCBackground;
    values.foreground = side->ui->fgcolor;
    values.background = side->ui->bgcolor;

    init_fonts(side);

    gc = XCreateGC(sdd(side), side->ui->game, mask, &values);
    side->ui->gc = XCreateGC(sdd(side), side->ui->game, mask, &values);
    side->ui->textgc = XCreateGC(sdd(side), side->ui->game, mask, &values);
    side->ui->icongc = XCreateGC(sdd(side), side->ui->game, mask, &values);
    /* for border drawing */
    values.line_width = 3;
    mask |= GCLineWidth;
    side->ui->varicongc = XCreateGC(sdd(side), side->ui->game, mask, &values);
    side->ui->unitgc = XCreateGC(sdd(side), side->ui->game, mask, &values);
    values.foreground = side->ui->bgcolor;
    values.background = side->ui->fgcolor;
    side->ui->invicongc = XCreateGC(sdd(side), side->ui->game, mask, &values);
    values.function = GXclear;
    mask = GCFunction;
    side->ui->cleargc = XCreateGC(sdd(side), side->ui->game, mask, &values);
    DGprintf("doing textgc...\n");
    XSetFont(sdd(side), side->ui->textgc, side->ui->textfont->fid);
    DGprintf("doing icongc ...\n");
    if (side->ui->iconfont != NULL) {
	XSetFont(sdd(side), side->ui->icongc, side->ui->iconfont->fid);
    }
    DGprintf("done fonts ...\n");
    mask = GCFillStyle | GCGraphicsExposures;
    values.fill_style = FillSolid;
    values.graphics_exposures = FALSE;
    XChangeGC(sdd(side), side->ui->unitgc, mask, &values);
    XChangeGC(sdd(side), side->ui->varicongc, mask, &values);
    if (side->ui->unitfont != NULL) {
	XSetFont(sdd(side), side->ui->unitgc, side->ui->unitfont->fid);
    }
    if (side->ui->iconfont != NULL) {
	XSetFont(sdd(side), side->ui->invicongc, side->ui->iconfont->fid);
	XSetFont(sdd(side), side->ui->varicongc, side->ui->iconfont->fid);
    }

    init_bitmaps(side);

    /* Since clearbitgc has to have a depth of 1 we'll */
    /* hang it off bombpics[0] */
    mask = GCFunction;
    values.function = GXclear;
    side->ui->clearbitgc = XCreateGC(sdd(side), side->ui->bombpics[0], mask, &values);

    init_cursors(side);

    XFreeGC(sdd(side), gc);
#endif
}

init_bitmaps(side)
Side *side;
{
#if 0
    side->ui->bombpics[0] = XCreateBitmapFromData(sdd(side), side->ui->game,
	bomb1_bits, bomb1_width, bomb1_height);
    side->ui->bombpics[1] = XCreateBitmapFromData(sdd(side), side->ui->game,
        bomb2_bits, bomb2_width, bomb2_height);
    side->ui->bombpics[2] = XCreateBitmapFromData(sdd(side), side->ui->game,
	bomb3_bits, bomb3_width, bomb3_height);
    side->ui->bombpics[3] = XCreateBitmapFromData(sdd(side), side->ui->game,
	bomb4_bits, bomb4_width, bomb4_height);
    side->ui->hitpics[0] = XCreateBitmapFromData(sdd(side), side->ui->game,
	miss_bits, miss_width, miss_height);
    side->ui->hitpics[1] = XCreateBitmapFromData(sdd(side), side->ui->game,
	hit_bits, hit_width, hit_height);
    side->ui->hitpics[2] = XCreateBitmapFromData(sdd(side), side->ui->game,
	kill_bits, kill_width, kill_height);
    side->ui->boxcurs = XCreateBitmapFromData(sdd(side), side->ui->game,
	boxcurs_bits, boxcurs_width, boxcurs_height);
    side->ui->boxmask = XCreateBitmapFromData(sdd(side), side->ui->game,
	boxmask_bits, boxmask_width, boxmask_height);
    DGprintf("Bitmaps stored ...\n");
#endif
}

#if 0
init_fonts(side)
Side *side;
{
    side->ui->textfont =
      open_font(side, TEXTFONT, "TextFont", NULL, "");
    side->ui->fw = font_width(side->ui->textfont);
    side->ui->fh = font_height(side->ui->textfont);
    side->ui->cursorfont =
      open_font(side, "cursor", "?", NULL, "");
    setup_unit_icons(side);
    setup_terrain_icons(side);
    {
	int p, u;
	XFontStruct *font;

	font = open_font(side, "*-bold-r-*-10-*", "LegendFont",
			 side->ui->textfont, "text");
	
	for (p = 0; p < NUMPOWERS; ++p) {
	    for_all_unit_types(u) {
		side->ui->ulegendfonts[p][u] = font;
	    }
	}
    }
}

/* Make all the different kinds of cursors we intend to use.  Should be
   one for each kind of mode or tool. */

init_cursors(side)
Side *side;
{
#if 0
    int i, w, h;

    for (i = 0; i < NUMCONTROLS; ++i) {
	side->ui->toolpics[i] = None;
    }
    for (i = 0; i < NUMTOOLS; ++i) {
	side->ui->toolcursors[i] = None;
    }
    rootcursor = FALSE;
    XQueryBestCursor(sdd(side), side->ui->game,
		     lookglass_width, lookglass_height, &w, &h);
    DGprintf("Best cursor shape is %dx%d\n", w, h);
    if (w >= lookglass_width && h >= lookglass_height) {
        side->ui->toolcursors[LOOKTOOL] =
	  make_cursor(sdd(side), side->ui->game,
		      lookglass_bits, lookmask_bits, white_color(side),
		      black_color(side), lookglass_x_hot, lookglass_y_hot);
	side->ui->toolpics[LOOK] =
	  XCreateBitmapFromData(sdd(side), side->ui->game,
				lookglass_bits, 16, 16);
        side->ui->toolcursors[UNITMOVETOOL] =
	  make_cursor(sdd(side), side->ui->game,
		      movecursor_bits, movemask_bits, white_color(side),
		      black_color(side), movecursor_x_hot, movecursor_y_hot);
	side->ui->toolpics[UNITMOVE] =
	  XCreateBitmapFromData(sdd(side), side->ui->game,
				movecursor_bits, 16, 16);
        side->ui->toolcursors[UNITSHOOTTOOL] =
	  make_cursor(sdd(side), side->ui->game,
		      shootcursor_bits, shootmask_bits, white_color(side),
		      black_color(side), shootcursor_x_hot, shootcursor_y_hot);
	side->ui->toolpics[UNITSHOOT] =
	  XCreateBitmapFromData(sdd(side), side->ui->game,
				shootcursor_bits, 16, 16);

        side->ui->toolcursors[HEXPAINTTOOL] =
	  make_cursor(sdd(side), side->ui->game,
		      hexcursor_bits, hexcursor_bits, white_color(side),
		      black_color(side), hexcursor_x_hot, hexcursor_y_hot);
	side->ui->toolpics[HEXPAINT] =
	  XCreateBitmapFromData(sdd(side), side->ui->game,
				hexcursor_bits, 16, 16);
        side->ui->toolcursors[BORDPAINTTOOL] =
	  make_cursor(sdd(side), side->ui->game,
		      bordcursor_bits, bordcursor_bits, white_color(side),
		      black_color(side), bordcursor_x_hot, bordcursor_y_hot);
	side->ui->toolpics[BORDPAINT] =
	  XCreateBitmapFromData(sdd(side), side->ui->game,
				bordcursor_bits, 16, 16);
        side->ui->toolcursors[CONNPAINTTOOL] =
	  make_cursor(sdd(side), side->ui->game,
		      conncursor_bits, conncursor_bits, white_color(side),
		      black_color(side), conncursor_x_hot, conncursor_y_hot);
	side->ui->toolpics[CONNPAINT] =
	  XCreateBitmapFromData(sdd(side), side->ui->game,
				conncursor_bits, 16, 16);
	side->ui->toolpics[UNITADD] =
	  XCreateBitmapFromData(sdd(side), side->ui->game,
				addunits_bits, 16, 16);

	side->ui->toolpics[ZOOMOUT] =
	  XCreateBitmapFromData(sdd(side), side->ui->game,
				farther_bits, 16, 16);
	side->ui->toolpics[ZOOMIN] =
	  XCreateBitmapFromData(sdd(side), side->ui->game,
				closer_bits, 16, 16);
	side->ui->toolpics[SHOWTERR] =
	  XCreateBitmapFromData(sdd(side), side->ui->game,
				terr_bits, 16, 16);
	side->ui->toolpics[SHOWTEMP] =
	  XCreateBitmapFromData(sdd(side), side->ui->game,
				therm_bits, 16, 16);
	side->ui->toolpics[SHOWUNITS] =
	  XCreateBitmapFromData(sdd(side), side->ui->game,
				units_bits, 16, 16);
	side->ui->toolpics[SHOWNAMES] =
	  XCreateBitmapFromData(sdd(side), side->ui->game,
				names_bits, 16, 16);
    } else {
	init_warning("Warning: Can't have %dx%d cursors on \"%s\"!",
		     lookglass_width, lookglass_height, side->player->displayname);
        init_warning("Using default cursor instead.");
        rootcursor = TRUE;
    }
    DGprintf("Cursors stored ...\n");
#endif
}

/* Since XCreatePixmapCursor() takes XColors and not pixel values we
   have to look up the colors associated with the foreground and
   background pixel values in the color table and pass them to
   XCreatePixmapCursor(). */
   
Cursor
make_cursor(dpy, win, cursbits, maskbits, fg, bg, x, y)
Display *dpy;
Window win;
char *cursbits, *maskbits;
unsigned long fg, bg;
unsigned int x, y;
{
#if 0
    Pixmap curs, mask;
    XColor fgcolor, bgcolor;
    Cursor rslt;

    mask = XCreateBitmapFromData(dpy, win, maskbits, 16, 16);
    curs = XCreateBitmapFromData(dpy, win, cursbits, 16, 16);
    fgcolor.pixel = fg;
    XQueryColor(dpy, DefaultColormap(dpy, DefaultScreen(dpy)), &fgcolor);
    bgcolor.pixel = bg;
    XQueryColor(dpy, DefaultColormap(dpy, DefaultScreen(dpy)), &bgcolor);
    rslt = XCreatePixmapCursor(dpy, curs, mask, &fgcolor, &bgcolor, x, y);
    XFreePixmap(dpy, mask);
    XFreePixmap(dpy, curs);
    return rslt;
#endif
}
#endif

/* Get a total bounding box on the font. */

font_width(font)
XFontStruct *font;
{
    return font->max_bounds.width;
}

font_height(font)
XFontStruct *font;
{
    return font->max_bounds.ascent + font->max_bounds.descent;
}

string_width(str, font)
char *str;
XFontStruct *font;
{
    return (strlen(str) * font_width(font));
}

/* Come up with font characters, bitmaps, or pixmaps for all the terrain types.
   This should find all versions and sizes and make them available. */

setup_terrain_icons(side)
Side *side;
{
#if 0
    int i, t, p;
    int junk1, junk2;
    int foundsome = FALSE;
    short w, h;
    Pixmap pat;
    int w2, h2;


    /* Get the solid hex outlines. */
    side->ui->hexpics[3] = XCreateBitmapFromData(sdd(side), side->ui->game,
	hex8_bits, hex8_width, hex8_height);
    side->ui->bhexpics[3] = XCreateBitmapFromData(sdd(side), side->ui->game,
	hex8b_bits, hex8b_width, hex8b_height);
    side->ui->hexpics[4] = XCreateBitmapFromData(sdd(side), side->ui->game,
	hex16_bits, hex16_width, hex16_height);
    side->ui->bhexpics[4] = XCreateBitmapFromData(sdd(side), side->ui->game,
	hex16b_bits, hex16b_width, hex16b_height);
    /* Now attempt to pick up any fonts that might be available. */
    /* (How to do more than one?) */
    side->ui->iconfont = open_font(side, ICONFONT, "IconFont", NULL, "");
    if (side->ui->iconfont != NULL) {
	/* See if the font matches any of the magnifications. */
	w = font_width(side->ui->iconfont);
	h = font_height(side->ui->iconfont);
	for (p = 0; p < NUMPOWERS; ++p) {
	    if (hws[p] == w && hhs[p] == h) {
		for_all_terrain_types(t) {
		    /* should also test for presence of each terrtype */
		    side->ui->usewhat[p][t] = usefontchars;
		    side->ui->terrchars[p][t] = t_type_char(t);
		    side->ui->terrfonts[p][t] = side->ui->iconfont;
		    foundsome = TRUE;
		}
	    }
	}
	if (!foundsome) {
	    init_warning("The font \"%s\" (size %dx%d) was useless",
			 ICONFONT, w, h);
	}
    } else {
	/* Be informative, especially for new X users. */
	init_warning("Xconq is usually much faster if fonts are installed.");
	init_warning("Please consult the docs and/or your site admins for details.");
	/* Mark this as undefined. */
	side->ui->iconfont = NULL;
    }
    /* Pick up any bitmaps for terrain that might be defined. */
    for_all_terrain_types(t) {
	if (t_pattern_name(t) && (strlen(t_pattern_name(t)) > 0)) {
	    pat = load_bitmap(side, t_pattern_name(t), "t", &w2, &h2);
	    if (pat != None) { 
		for (p = 0; p < NUMPOWERS; ++p) {
		    if (hws[p] >= w2 && hhs[p] >= h2) {
			side->ui->terrpics[p][t] = pat;
			/* Forget "usewhat" and detect directly? */
			if (side->ui->usewhat[p][t] == usefontchars) {
			    side->ui->usewhat[p][t] = usepictures;
			}
		    }
		}
	    } else {
		/* nothing to do? */
	    }
	} else {
	    set_t_pattern_name(t, NULL);
	}
    }
#endif
}

/* Make pixmap tiles for large hexes by converting font chars. */

setup_terrain_tiles(side)
Side *side;
{
#if 0
    char ch;
    int p, t, w = 0, h = 0, p0, sx, sy;

    for_all_terrain_types(t) {
	for (p = 0; p < NUMPOWERS; ++p) {
	    if (w == 0 && side->ui->usewhat[p][t] == usefontchars) {
		p0 = p;
		w = hws[p0];  h = hcs[p0];
	    } else if (w > 0) {
		/* Transform the font char into a tiling pattern. */
		Pixmap tile = XCreatePixmap(sdd(side), side->ui->game,
					    w, h, 1);
		ch = side->ui->terrchars[p0][t];
		sx = 0;  sy = side->ui->iconfont->max_bounds.ascent;
		XSetForeground(sdd(side), side->ui->icongc, black_color(side));
		XSetBackground(sdd(side), side->ui->icongc, black_color(side));
		XFillRectangle(sdd(side), tile, side->ui->icongc,
			       0, 0, w, h);
		XSetForeground(sdd(side), side->ui->icongc, white_color(side));
		XDrawString(sdd(side), tile, side->ui->icongc,
			    sx, sy, &ch, 1);
		XDrawImageString(sdd(side), tile, side->ui->icongc,
				 sx - w / 2, sy - h, &ch, 1);
		XDrawImageString(sdd(side), tile, side->ui->icongc,
				 sx + w / 2, sy - h, &ch, 1);
		side->ui->terrpics[p][t] = tile;
	    }
	}
    }
#endif
}

t_type_char(t) int t; { return '?'; }
u_type_char(u) int u; { return '?'; }

/* Come up with font characters, bitmaps, or pixmaps for all the unit types,
   including sponging off the text font. */

setup_unit_icons(side)
Side *side;
{
#if 0
    int foundsome = FALSE;
    short sw, sh;
    int p, u, w, h;
    Pixmap pic;

    /* (Should be able to skip font loading if units don't need it) */
    side->ui->unitfont = open_font(side, "standard", "UnitFont",
				    side->ui->textfont);
    sw = font_width(side->ui->unitfont);
    sh = font_height(side->ui->unitfont);
    /* shouldn't set any of this up unless it knows that the font & char
       are the right ones, at whatever power. */
#if 0
    /* See if the font is usable at any of the magnifications. */
    for (p = 0; p < NUMPOWERS; ++p) {
	/* OK to use small icons on big hexes. */
	if (uws[p] >= sw && uhs[p] >= sh) {
	    for_all_unit_types(u) {
		side->ui->unitchars[p][u] = u_type_char(u);
		side->ui->unitfonts[p][u] = side->ui->unitfont;
		side->ui->unitw[p][u] = sw;  side->ui->unith[p][u] = sh;
		foundsome = TRUE;
	    }
	}
    }
#endif
    if (!foundsome) {
	init_warning("The font \"%s\" (size %dx%d) was useless",
		     "standard", sw, sh);
    }
    /* Individual icon names override font chars (junky) */
    for_all_unit_types(u) {
	if (u_icon_name(u) && (strlen(u_icon_name(u)) > 0)) {
	    pic = load_bitmap(side, u_icon_name(u), "u", &w, &h);
	    if (pic != None) { 
		for (p = 0; p < NUMPOWERS; ++p) {
		    if (uws[p] >= w && uhs[p] >= h) {
			side->ui->unitpics[p][u] = pic;
			/* Flush any char that might have been set. */
			side->ui->unitchars[p][u] = '\0';
			side->ui->unitfonts[p][u] = NULL;
			side->ui->unitw[p][u] = w;  side->ui->unith[p][u] = h;
		    }
		}
	    } else {
		/* Making sure */
		set_u_icon_name(u, NULL);
	    }
	} else {
	    /* Making sure */
	    set_u_icon_name(u, NULL);
	}
    }
    /* Try to make larger/smaller derived icons. */
    for_all_unit_types(u) {
	for (p = 1; p < NUMPOWERS; ++p) {
	    if (side->ui->unitpics[p-1][u] != None
		&& side->ui->unitpics[p][u] == None) {
		/* Next smaller icon is defined, expand it. */
	    }
	}
	for (p = 0; p < NUMPOWERS-1; ++p) {
	    if (side->ui->unitpics[p][u] == None
		&& side->ui->unitpics[p+1][u] != None) {
		/* Next bigger icon is defined, shrink it. */ 
	    }
	}
    }
#endif
}

/* Given a magnification power, update the map to reflect this. */

set_map_power(mw)
MapWidget mw;
{
    int power = mw->map.power, upower;

    mw->map.mag = mags[power];
    mw->map.hw = hws[power];  mw->map.hh = hhs[power];
    mw->map.hch = hcs[power];
    /* this should be smarter somehow */
    upower = mw->map.power;
    if (upower > 4) upower = 4;
    mw->map.uw = uws[upower];  mw->map.uh = uhs[upower];
    /* could work back from pixels to come up with exactly
       the right viewport size. */
}

/* Given a map with already-set pixel size and viewport center, compute
   reasonable values for its size and position in world coordinates.
   This is really just caching, can be run at any time. */

set_map_viewport(mw)
MapWidget mw;
{
    Dprintf("Setting map viewport\n");
    /* Compute a viewport width and height from size in pixels. */
    mw->map.vw = min(area.width, mw->core.width / mw->map.hw);
    /* what about the hh - hch correction here? */
    mw->map.vh = min(area.height, mw->core.height / mw->map.hch);
    /* Compute some cached values that are used a lot (still?). */
    mw->map.vw2 = mw->map.vw / 2;  mw->map.vh2 = mw->map.vh / 2;
    /* Since we don't want to let the viewport stare into the void
       too much, the code here keeps the position limited to reasonable
       values, *except* that the left and right sides of a hexagon may be
       exposed, so other code has to make sure that nonsensical positions
       are not displayed. */
    if (mw->map.isworld) {
	/* World maps are simple, position is fixed. */
	mw->map.vx = (area.xwrap ? 0 : area.height / 4);
	mw->map.vy = 0;
    } else {
	/* Set the "normal" value of the position of the viewport. */
	mw->map.vx = mw->map.vcx - mw->map.vw2;
	mw->map.vy = mw->map.vcy - mw->map.vh2;
	/* Now move the viewport in so it doesn't go outside the area. */
	if (area.xwrap) {
	    if (mw->map.vx < 0)
	      mw->map.vx = (mw->map.vx + 10 * area.width) % area.width;
	    /* can't possibly get too large */
	} else {
	    if (mw->map.vx < 0)
	      mw->map.vx = 0;  /* not quite right? */
	    if (mw->map.vx > (area.width - mw->map.vw))
	      mw->map.vx = area.width - mw->map.vw;
	}
	if (mw->map.vy < 0)
	  mw->map.vy = 0;
	if (mw->map.vy > area.height - mw->map.vh)
	  mw->map.vy = area.height - mw->map.vh;
    }
}

/* Return the pixel values for black and white. */

white_color(side)
Side *side;
{
    return WhitePixel(sdd(side), DefaultScreen(sdd(side)));
}

black_color(side)
Side *side;
{
    return BlackPixel(sdd(side), DefaultScreen(sdd(side)));
}

/* Get a color set up and warn if not getting what was asked for.  We can */
/* tolerate being off somewhat.  (Note X rgb value range is 0-65535.) */

#define CLOSE_ENOUGH(X,Y) (abs(((int) X) - ((int) Y)) < 2000)

long
request_color(side, name)
Side *side;
char *name;
{
    XColor c, avail;

    DGprintf("Requesting color %s\n", name);
    /* This might be called to get user-specified colors, even on a mono
       display, so deal with it. */
    if (name == NULL || strlen(name) == 0) {
	init_warning(
          "Requesting unnamed color on display \"%s\", substituting white",
		     side->player->displayname);
	return white_color(side);
    } else if (XDisplayCells(sdd(side), DefaultScreen(sdd(side))) <= 2) {
	if (strcmp("white", name) == 0) {
	    return white_color(side);
	} else if (strcmp("black", name) == 0) {
	    return black_color(side);
	} else {
	    init_warning(
              "No color \"%s\" on the mono display \"%s\", substituting white",
			 name, side->player->displayname);
	    return white_color(side);
	}
    } else {
	XAllocNamedColor(sdd(side),
			 DefaultColormap(sdd(side), DefaultScreen(sdd(side))),
			 name, &avail, &c);
	if (!(CLOSE_ENOUGH(c.red, avail.red)
	      && CLOSE_ENOUGH(c.green, avail.green)
	      && CLOSE_ENOUGH(c.blue, avail.blue))) {
	    init_warning("Warning: %s color is way off on display \"%s\"!",
			 name, side->player->displayname);
	    init_warning("(%d %d %d instead of %d %d %d), but using it anyway",
			 avail.red, avail.green, avail.blue,
			 c.red, c.green, c.blue);
	}
	return avail.pixel;
    }
}

/* Acquire a set of colors. */

init_colors(side)
Side *side;
{
#if 0
    DGprintf("%d colors available ...\n", display_colors(side));
    side->ui->monochrome = (display_colors(side) <= 2);
    set_colors(side);
#endif
}

/* This will set up the correct set of colors at any point in the game.
   Colors are all specified by name; if any are not available, it is up to
   the graphics interface to supply a substitute. */

set_colors(side, w)
Side *side;
Widget w;
{
    int	t;
    long fg, bg;
    char *colorname;
    MapWidget mw = (MapWidget) w;

    if (XDisplayCells(sdd(side), DefaultScreen(sdd(side))) <= 2) {
	for_all_terrain_types(t) {
	    /* or some such */
	    mw->map.hexcolor[t] = mw->core.background_pixel;
	}
	/* etc */
    } else {
	for_all_terrain_types(t) {
	    colorname = t_color(t);
	    /* Warn if we have to substitute a valid color name. */
	    /* (Do this in generic period checking?) */
	    if (colorname == NULL || strlen(colorname) == 0) {
		init_warning(
	      "Warning: no color for terrain type \"%s\", substituting white",
			     t_type_name(t));
		colorname = "white";
	    }
	    mw->map.hexcolor[t] = request_color(side, colorname);
	}
#if 0
	if (!XmuDistinguishablePixels(sdd(side), ?, mw->map.hexcolor, numttypes)) {
	    /* generate a warning */
	}
#endif
#if 0
	side->ui->owncolor = request_color(side, "black");
	side->ui->altcolor = request_color(side, "blue");
	side->ui->diffcolor = request_color(side, "maroon");
	side->ui->bdcolor = request_color(side, "blue");
	side->ui->graycolor = request_color(side, "light gray");
	side->ui->enemycolor = request_color(side, "red");
	side->ui->neutcolor = request_color(side, "gray");
	side->ui->goodcolor = request_color(side, "green");
	side->ui->badcolor = request_color(side, "red");
#endif
    }
}

/* Map legends are a special kind of popup widget (eventually). */

/* Generate a legend for the pictures of terrain and units. */

/* popup should work from a power derived from map's current power,
   perhaps resize as map zooms in and out?, should consist of imlists
   for both terrains and units. */

legend_help(side)
Side *side;
{
#if 0
    int t, u, dir, tcol = 3 * side->ui->margin + hws[4];
    int spacing = hhs[4] + 4 * side->ui->margin;
    int offset = hhs[4]/4, second = 20*10;

    second += 2 * (side->ui->margin + hws[4]);
    tcol += 2 * (side->ui->margin + hws[4]);
    /* List all the terrain patterns/colors. */
    /* (All these should have bgs or be outlined or something) */
    for_all_terrain_types(t) {
	if (t_can_be_hex(t)) {
	    draw_hex_icon(side, side->ui->help,
			  side->ui->margin, t * spacing, t, 4,
			  (side->ui->monochrome ? side->ui->fgcolor :
			   side->ui->hexcolor[t]));
	}
	/* (These should be offset so we can see the types that can have
	   multiple roles.) */
	if (t_can_be_border(t)) {
	    for_all_directions(dir) {
		draw_border_line(side, side->ui->help,
				 2 * side->ui->margin + hws[4],
				 t * spacing, dir,
				 4, side->ui->hexcolor[t]);
	    }
	}
	if (t_can_be_connection(t)) {
	    for_all_directions(dir) {
		draw_connection_line(side, side->ui->help,
				     3 * side->ui->margin + 2 * hws[4],
				     t * spacing, dir,
				     4, side->ui->hexcolor[t]);
	    }
	}
	draw_fg_text(side, side->ui->help,
		     tcol, t * spacing + offset, t_type_name(t));
    }
    /* List all the unit types. */
    for_all_unit_types(u) {
	/* (Need to do at preferred magnification) */
	draw_unit_icon(side, side->ui->help,
		       second, u * spacing, u, 4,
		       side->ui->fgcolor, side->ui->bgcolor);
	sprintf(tmpbuf, "%s (%s)", u_type_name(u), u_short_name(u));
	draw_fg_text(side, side->ui->help,
		     tcol + second, u * spacing + offset, tmpbuf);
    }
#endif
}
