/* Copyright (c) 1987, 1988, 1991, 1992  Stanley T. Shebs. */
/* This program may be used, copied, modified, and redistributed freely */
/* for noncommercial purposes, so long as this notice remains intact.  */

/* Interface implementations for the X11 version of xconq. */

#include "conq.h"
#include "xtconq.h"

extern int margin;
extern int bd;
extern int conwid;
extern int sbarwid;

Side *
side_from_display(dpy)
Display *dpy;
{
    Side *side;

    for_all_sides(side) {
	if (active_display(side) && sdd(side) == dpy) {
	    return side;
	}
    }
    return NULL;
}

/* Look at a single event and fill the request structure appropriately.
   The "side" might be a indepside for init, so watch for this. */

/* (This needs to check on timeouts?) */

process_event(side, evt)
Side *side;
XEvent evt;
{
    char buf[BUFSIZE], ch;
    int nchar, rawx, rawy;
    Map *map;

    DGprintf("Display %s gets an event\n", side->player->displayname);
#if 0
    switch (evt.type) {
      case KeyPress:
	nchar = XLookupString(&(evt.xkey), buf, BUFSIZE, NULL, NULL);
	if (nchar > 0) {
	    ch = buf[0];
	    DGprintf("Display %s returns char '%c'\n", side->player->displayname, ch);
	    /* so we can do ^L even while typing a string */
	    if (ch == '\014') { /* control l */
		redraw(side);
		/* does help win too? */
	    } else if (evt.xbutton.window == side->ui->game) {
		/* make more sophisticated eventually */
/*		reveal_window(side, side->ui->ourcloseup->closeupwin);  */
	    } else if (evt.xbutton.window == side->ui->ourcloseup->closeupwin) {
	    } else {
		side->ui->reqch = ch;
		/* Call the handler, giving it side and cancel flag. */
		if (side->ui->modalhandler) {
		    int (*fn)() = side->ui->modalhandler;

		    /* Take out the handler, will restore itself if needed. */
		    side->ui->modalhandler = NULL;
		    (*fn)(side, (ch == '\033'));
		} else {
		    x_command(side);
		}
	    }
	} else {
	    /* pretty strange, ok to do nothing? */
	}
	break;
      case ButtonPress:
	if (button_down_map(side, evt)) {
	} else if (button_down_list(side, evt)) {
	} else if (button_down_closeup(side, evt)) {
	} else if (evt.xbutton.window == side->ui->msg) {
	    /* try scrolling the msg window */
	} else if (evt.xbutton.window == side->ui->ourcloseup->closeupwin) {
	    button_down_docdial(side, evt.xbutton.x, evt.xbutton.y);
	} else {
	    DGprintf("Button press on unhandled window\n");
	}
	break;
      case ButtonRelease:
	if (button_up_map(side, evt)) {
	} else if (button_up_list(side, evt)) {
	} else if (button_up_closeup(side, evt)) {
	} else if (evt.xbutton.window == side->ui->msg) {
	} else if (evt.xbutton.window == side->ui->ourcloseup->closeupwin) {
	} else {
	    DGprintf("Button release on unhandled window\n");
	}
	break;
      case Expose:
	/* Handle the expose events.  This code needs to be careful not to
	   do too many redraws, since some of them can be quite slow. */
	do {
	    if (evt.xexpose.window == side->ui->msg) {
		show_note(side);
	    } else if (evt.xexpose.window == side->ui->game) {
		show_game_state(side);
	    } else if (evt.xexpose.window == side->ui->ourcloseup->closeupwin) {
	    } else if (expose_map(side, evt)) {
	    } else if (expose_list(side, evt)) {
	    } else if (expose_closeup(side, evt)) {
	    } else {
		DGprintf("Exposure of unhandled window\n");
	    }
	} while (XCheckMaskEvent(sdd(side), ExposureMask, &evt));
	/* We need to blast *some* exposure events here, but not all... */
	break;
      case DestroyNotify:
	run_warning("want to destroy a window");
	break;
      case ConfigureNotify:
	/* Need to recalculate viewports if map windows change size. */
	for_all_maps(side, map) {
	    if (evt.xconfigurerequest.window == map->window) {
		int sbaradj = (map->isworld ? 0 : sbarwid);

		break;
	    }
	}
	/* Other win types don't need anything special, just fall through. */
	break;
      default:
	/* X might send us events we don't care about, so don't get bent
	   out of shape. */
        run_warning("Unknown X event type %d, ignoring", evt.type);
	break;
    }
#endif
    return FALSE;
}

#if 0
    /* some commands OK, so where should this go? */
    if (period_giventime() && !side->timedout && side->timeleft <= 0) {
	side->timedout = TRUE;
	notify(side, "You ran out of time!!");
	beep(side);
    }
#endif

void
done_utype_dialog(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));

    if (side) {
	/* should extract a selected type */
	side->ui->reqch = NONUTYPE;
	XtPopdown(sd(side)->utypedialog);
    }
}


void
cancel_utype_dialog(w, clientdata, calldata)
Widget w;
XtPointer clientdata, calldata;
{
    Side *side = side_from_display(XtDisplay(w));

    if (side) {
	side->ui->reqch = NONUTYPE;
	XtPopdown(sd(side)->utypedialog);
    }
}

char *utypenames[100];

/* Prompt for a type of a unit from player, maybe only allowing some types
   to be accepted.  Also allow specification of no unit type.  We do this
   by scanning the vector, building a string of chars and a vector of
   unit types, so as to be able to map back when done. */

ask_unit_type(side, prompt, possibles)
Side *side;
char *prompt;
int *possibles;
{
    int i, numtypes = 0, u, type;

    for_all_unit_types(u) {
	if (possibles == NULL || possibles[u]) {
	    utypenames[numtypes++] = u_type_name(u);
	}
    }
    if (numtypes == 0) {
	type = NONUTYPE;
    } else {
	if (sd(side)->utypedialog == NULL) {
	    Widget utypebox, utypelist, done, cancel;

	    i = 0;
	    XtSetArg(tmpargs[i], XtNx, 50);  ++i;
	    XtSetArg(tmpargs[i], XtNy, 50);  ++i;
	    sd(side)->utypedialog =
	      XtCreatePopupShell("unit type dialog", overrideShellWidgetClass,
				 toplevel(side), tmpargs, i);
	    utypebox =
	      XtCreateManagedWidget("box", boxWidgetClass,
				    sd(side)->utypedialog, (Arg *) NULL, 0);
	    i = 0;
	    XtSetArg(tmpargs[i], XtNlabel, prompt);  ++i;
	    XtCreateManagedWidget("query", labelWidgetClass,
				  utypebox, tmpargs, i);
	    i = 0;
	    XtSetArg(tmpargs[i], XtNlist, utypenames);  ++i;
	    XtSetArg(tmpargs[i], XtNnumberStrings, numtypes);  ++i;
	    XtSetArg(tmpargs[i], XtNverticalList, False); ++i;
	    XtSetArg(tmpargs[i], XtNdefaultColumns, 1); ++i;
	    utypelist =
	      XtCreateManagedWidget("unit type list", listWidgetClass,
				    utypebox, tmpargs, i);
	    i = 0;
	    XtSetArg(tmpargs[i], XtNlabel, "OK");  ++i;
	    done =
	      XtCreateManagedWidget("command", commandWidgetClass,
				    utypebox, tmpargs, i);
	    XtAddCallback(done, XtNcallback, done_utype_dialog, NULL);
	    i = 0;
	    XtSetArg(tmpargs[i], XtNlabel, "Cancel");  ++i;
	    cancel =
	      XtCreateManagedWidget("command", commandWidgetClass,
				    utypebox, tmpargs, i);
	    XtAddCallback(cancel, XtNcallback, cancel_utype_dialog, NULL);
	    }
	/* plug in a unit list? */
	XtPopup(sd(side)->utypedialog, XtGrabNone);
	type = -1;
    }
    return type;
}

/* Do something with the char or unit type that the player entered. */

grok_unit_type(side)
Side *side;
{
    int i, type = -1;

    type = side->ui->reqch;

    return type;
}

ask_terrain_type(side, prompt, possibles)
Side *side;
char *prompt;
short *possibles;
{
    int numtypes = 0, t, type;

    for_all_terrain_types(t) {
	side->ui->bvec[t] = 0;
	if (possibles == NULL || possibles[t]) {
	    side->ui->bvec[t] = 1;
	    side->ui->uvec[numtypes] = t;
	    numtypes++;
	}
    }
    if (numtypes == 0) {
	type = NONUTYPE;
    } else if (numtypes == 1) {
	type = side->ui->uvec[0];
	side->ui->bvec[type] = 0;
    } else {
#if 0
	sprintf(side->ui->promptbuf, "%s [%s] ", prompt, side->ui->ustr);
	    reveal_prompt(side);
	show_prompt(side);
/*	draw_terrain_list(side, side->ui->bvec);  */
	type = -1;
#endif
    }
    return type;
}

/* Do something with the char or terrain type that the player entered. */

grok_terrain_type(side)
Side *side;
{
#if 0
    int i, type = -1;

    echo_at_prompt(side, side->ui->reqch);
    if (side->ui->reqch == '?') {
	/*	    help_terrain_type(side);  */
    } else if (side->ui->reqch == ESCAPE) {
	type = NONUTYPE;
#if 0
    } else if (iindex(side->ui->reqch, side->ui->ustr) == -1) {
	notify(side, "Terrain type '%c' not in \"%s\"!",
	       side->ui->reqch, side->ui->ustr);
#endif
    } else {
	/*	    type = find_terr(side->ui->reqch);  fix eventually */
    }
    if (type >= 0) {
	clear_prompt(side);
	for_all_terrain_types(i) side->ui->bvec[i] = 0;
    }
    return type;
#endif
}

/* User is asked to pick a position on map.  This will iterate until the */
/* space bar designates the final position. */

ask_position(side, prompt)
Side *side;
char *prompt;
{
    sprintf(side->ui->promptbuf,
	    "%s [mouse or keys to move, space bar to set]",
	    prompt);
}

/* Restore the saved "cur" slots. */

restore_cur(side)
Side *side;
{
}

/* Interpret the user's input in response to a position request.  All we */
/* have to comprehend is direction keys and mouse hits.  Space bars and */
/* unmoving mice both mean that a position has been decided on. */

grok_position(side)
Side *side;
{
    return FALSE;
}

Widget askyesno = NULL;
 
/* Prompt for a yes/no answer with a settable default. */

ask_bool(side, question, dflt)
Side *side;
char *question;
int dflt;
{
    char tmpbuf[100];

    sprintf(tmpbuf, "%s [%s]", question, (dflt ? "yn" : "ny"));

#if 0
    if (askyesno == NULL) {
	askyesno = XtCreateManagedWidget("box", boxWidgetClass,
					 thistoplevel, );
	/* make the buttons etc */
    }
#endif					 
    side->ui->reqvalue2 = dflt;
}

/* Figure out what the answer actually is, keeping the default in mind. */

grok_bool(side)
Side *side;
{
    int dflt = side->ui->reqvalue2;
#if 0
    char ch = side->ui->reqch;

    if (dflt ? (lowercase(ch) == 'n') : (lowercase(ch) == 'y'))
      dflt = !dflt;
#endif
    return dflt;
}

/* Read a string from the prompt window.  Deletion is allowed, and a */
/* cursor is displayed (this should definitely be a toolkit call...) */
/* Some restrictions on what strings can be read - for instance can't */
/* read or default to a NULL string. */

ask_string(side, prompt, dflt)
Side *side;
char *prompt, *dflt;
{
    if (dflt == NULL) {
	sprintf(side->ui->promptbuf, "%s ", prompt);
    } else {
	sprintf(side->ui->promptbuf, "%s (default \"%s\") ", prompt, dflt);
    }
    side->ui->reqcurstr = 0;
/*    write_str_cursor(side);  */
    side->ui->reqdeflt = dflt;
}

/* Dig a character from the filled-in request and add it into the string. */
/* Keep returning NULL until we get something. */

char *
grok_string(side)
Side *side;
{
    if (side->ui->reqch == '\r' ||
	side->ui->reqch == '\n' ||
	side->ui->reqcurstr >= BUFSIZE-2) {
	if (side->ui->reqcurstr == 0 && side->ui->reqdeflt != NULL) {
	    strcpy(side->ui->promptbuf, side->ui->reqdeflt);
	} else {
	    /* should get from a stringbuf or something */
	    (side->ui->promptbuf)[side->ui->reqcurstr] = '\0';
	}
	return copy_string(side->ui->promptbuf);
    } else {
	if (side->ui->reqch == BACKSPACE) {
	    if (side->ui->reqcurstr > 0) --side->ui->reqcurstr;
	} else {
/*	    echo_at_prompt(side, side->ui->reqch);  */
	    (side->ui->promptbuf)[side->ui->reqcurstr++] = side->ui->reqch;
	}
    }
/*    write_str_cursor(side);  */
    (side->ui->promptbuf)[side->ui->reqcurstr+1] = '\0';
    return NULL;
}

/* Open or reuse a unit closeup. */

make_current(side, unit)
Side *side;
Unit *unit;
{
    Map *map;

    if (active_display(side) && unit != NULL) {
	if (alive(unit)) {
	    if (side->ui->closeups == NULL) {
/*		create_closeup(side, unit, TRUE);  */
	    }
	    if (side->ui->closeups == NULL) {
		fprintf(stderr, "No closeup made!\n");
		return;
	    } else {
		/* not quite right */
		side->ui->closeups->unit = unit;
		show_closeup(side, side->ui->closeups);
	    }
	    /* a little too drastic probably */
	    for_all_maps(side, map) {
		map->curunit = unit;
		show_current_info(side, map);
	    }
	} else {
	    beep(side);
	}
    }
}

/* When doing a survey, you can move the cursor anywhere and it will show */
/* what is at that hex, but not give away too much! */

move_look(side, map, nx, ny)
Side *side;
Map *map;
int nx, ny;
{
    Unit *unit;
    Closeup *closeup;

    if (inside_area(nx, ny)) {
	erase_cursor(side, map);
	if (!in_view(side, nx, ny)) beep(side);
	unit = unit_at(nx, ny);
	if (unit != NULL && (unit->side == side || side->designer)) {
	    set_map_curunit(side, map, unit);
	    for_all_closeups(side, closeup) {
		if (closeup->unit == unit) {
		    /* bring to front */
		    /* flag on map(s) somehow? */
		    return;
		}
	    }
	    if (0 /* can add new window */) {
/*		create_closeup(side, unit);  */
	    } else if (side->ui->closeups) {
		/* reuse */
		side->ui->closeups->unit = unit;
	    }
	    show_unit_closeups(side, unit);
	} else {
	    map->curunit = NULL;
	    show_current_info(side, map);
	}
    } else {
	if (active_display(side)) beep(side);
    }
}

/* Set the "current" unit of a side - the one being displayed, moved, etc. */

set_map_curunit(side, map, unit)
Side *side;
Map *map;
Unit *unit;
{
    if (active_display(side)) {
	if (unit != NULL
	    && alive(unit)
	    && (unit->side == side || allied_side(unit->side,side) || Debug || numdesigners > 0)) {
	    map->curunit = unit;
	    show_current_info(side, map);
	} else {
	    map->curunit = NULL;
	    erase_cursor(side, map);
	    show_current_info(side, map);  /* maybe not? */
	}
    }
}

show_current_info(side, map)
Side *side;
Map *map;
{
    draw_cursor(side, map);
    if (map->curunit) {
	show_unit_closeups(side, map->curunit);
    }
}

/* See if the button event happened in a map display, record where etc,
   then return true. */

button_down_map(side, evt)
Side *side;
XEvent evt;
{
    Map *map;
    int rawx, rawy;

    for_all_maps(side, map) {
	if (evt.xbutton.window == (Window) map->viewwin) {
	    rawx = evt.xbutton.x;  rawy = evt.xbutton.y; 
	    /* Remember which map this happened on. */
	    side->ui->downmap = map;
	    /* Get world coords and closest hex side. */
	    deform(side, map, rawx, rawy,
		   &(side->ui->downx), &(side->ui->downy),
		   &(side->ui->downdir));
	    DGprintf("Display %s returns map down %d %d dir %d\n",
		     side->player->displayname,
		     side->ui->downx, side->ui->downy, side->ui->downdir);
	    return TRUE;
	}
    }
    return FALSE;
}

/* See if the button event happened in a map display, record where etc,
   then return true. */

button_up_map(side, evt)
Side *side;
XEvent evt;
{
    Map *map;
    int rawx, rawy, xx;

    for_all_maps(side, map) {
	if (evt.xbutton.window == (Window) map->viewwin) {
	    side->ui->upmap = map;
	    if (side->ui->upmap == side->ui->downmap) {
		rawx = evt.xbutton.x;  rawy = evt.xbutton.y; 
		/* Get world coords and closest hex side. */
		deform(side, map, rawx, rawy,
		       &(side->ui->upx), &(side->ui->upy),
		       &(side->ui->updir));
		DGprintf("Display %s returns map 0x%x up %d %d dir %d\n",
			 side->player->displayname, map,
			 side->ui->upx, side->ui->upy, side->ui->updir);
		if (side->ui->downx == side->ui->upx
		    && side->ui->downy == side->ui->upy) {
		    handle_map_click(side, map);
		} else {
		    handle_map_drag(side, map);
		}
	    } else {
		/* dragged to random place, forget about it */
	    }
	    return TRUE;
	}
    }
    return FALSE;
}

/* Clicking on a tool selects and changes the current tool. */

clicked_tool(side, map, which)
Side *side;
Map *map;
int which;
{
#if 0
    int i;

    for (i = 0; i < NUMCONTROLS; ++i) {
	if (controltypes[i].tool == map->curtool) {
	    map->controlb[i].selected = FALSE;
	    break;
	}
    }
    map->curtool = controltypes[which].tool;
    map->controlb[which].selected = TRUE;
    set_tool_cursor(side, map);
    return TRUE;
#endif
}

/* If a given tool (mode) has a special cursor, use it, otherwise go back
   to the look tool's cursor. */

set_tool_cursor(side, map)
Side *side;
Map *map;
{
#if 0
    int tool;

    if (sd(side)->toolcursors[map->curtool] != None) {
	tool = map->curtool;
    } else {
	/* revert to default cursor */
	tool = LOOKTOOL;
    }
    XDefineCursor(sdd(side), map->viewwin, sd(side)->toolcursors[tool]);
#endif
}

/* If clicked on the terrain type display, go to the next type. */
/* (should have something to go back too) */

clicked_ttype(side, map, which)
Side *side;
Map *map;
int which;
{
#if 0
    int t;

    switch (map->curtool) {
#ifdef LINEAR
      case BORDPAINTTOOL:
	t = map->curbordtype + 1;
	while (t != map->curbordtype) {
	    if (t >= numttypes) {
		t = -1; /* noborder */
		break;
	    } else if (t_can_be_border(t)) {
		break;
	    }
	    ++t;
	}
	map->curbordtype = t;
	break;
      case CONNPAINTTOOL:
	t = map->curconntype + 1;
	while (t != map->curconntype) {
	    if (t >= numttypes) {
		t = -1; /* noconnection */
		break;
	    } else if (t_can_be_connection(t)) {
		break;
	    }
	    ++t;
	}
	map->curconntype = t;
	break;
#endif /* LINEAR */
      case HEXPAINTTOOL:
	t = map->curttype + 1;
	/* This will loop forever if there are no valid hex types? */
	while (t != map->curttype) {
	    if (t >= numttypes) {
		t = 0;
		break;
	    } else if (t_can_be_hex(t)) {
		break;
	    }
	    ++t;
	}
	map->curttype = t;
	break;
      default:
	/* If any other tool in use, get to look at all ttypes. */
	++map->curttype;
	if (!between(0, map->curttype, numttypes-1)) map->curttype = 0;
	t = map->curttype;
	break;
    }
#if 0
    if (between(0, t, numttypes-1)) {
	map->controlb[which].color = side->ui->hexcolor[t];
    } else {
	map->controlb[which].color = side->ui->diffcolor;
    }
#endif
    /* Should do pic also, esp for noxxx cases */
    /* could do meaningful pics for linears also */
    return TRUE;
#endif
}

clicked_utype(side, map, which)
Side *side;
Map *map;
int which;
{
#if 0
    ++map->curutype;
    if (!between(0, map->curutype, numutypes-1)) map->curutype = 0;
    /* fails if units are in font chars or are wrong size... */
    map->controlb[which].pic = sd(side)->unitpics[4][map->curutype];
    return TRUE;
#endif
}

/* Cycle around possible sides, being sure to include "indep" as side 0. */

clicked_curside(side, map, which)
Side *side;
Map *map;
int which;
{
#if 0
    ++map->curside;
    if (!between(0, map->curside, numsides)) map->curside = 0;
    /* do emblems instead? */
    sprintf(map->controlb[which].text, "%d", map->curside);
    return TRUE;
#endif
}

/* Clicked on a zoom button. */

clicked_zoom(side, map, which)
Side *side;
Map *map;
int which;
{
}

/* Clicking on a display style usually toggles it. */

clicked_style(side, map, which)
Side *side;
Map *map;
int which;
{
#if 0
    int state;

    switch (which) {
      case SHOWTERR:
	state = map->drawterrain = !map->drawterrain;
	break;
      case SHOWPATS:
	state = map->drawhexpats = !map->drawhexpats;
	break;
      case SHOWEDGES:
	state = map->drawhexedges = !map->drawhexedges;
	break;
      case SHOWUNITS:
	state = map->drawunits = !map->drawunits;
	break;
      case SHOWNAMES:
	state = map->drawnames = !map->drawnames;
	break;
      case SHOWPEOPLE:
	state = map->drawpeople = !map->drawpeople;
	break;
      case SHOWELEVS:
	state = map->drawelevations = !map->drawelevations;
	break;
      case SHOWTEMP:
	state = map->drawtemp = !map->drawtemp;
	break;
      default:
	break;
    }
    /* Tweak the control to reflect the correct internal state. */
#if 0
    map->controlb[which].selected = state;
#endif
    /* also turn off others? */
    /* display the effect */
    show_map(side, map, FALSE, FALSE);
    return FALSE;
#endif
}

/* Invoke the appropriate action for a control panel item. */

handle_control_click(side, map, x, y)
Side *side;
Map *map;
int x, y;
{
}

/* Put the map at an absolute position. */

scroll_map_absolute(side, map, sx, sy)
Side *side;
Map *map;
int sx, sy;
{
#if 0
    /* World maps don't scroll. */
    if (map->isworld) return;
    undraw_box(side, map, side->ui->world);
    if (sx >= 0) {
	/* Only do if viewport is smaller than world. */
	if (area.xwrap || map->vw < area.width) {
	    map->vcx = ((area.width * sx) / map->pxw);
	}
    }
    if (sy >= 0) {
	if (map->vh < area.height) {
	    map->vcy = (((area.height * (map->pxh - sy)) / map->pxh));
	}
    }
    /* Now adjust the viewport position as needed. */
    set_map_viewport(map);
    show_map(side, map, FALSE, TRUE);
#endif
}

handle_map_click(side, map)
Side *side;
Map *map;
{
#if 0
    if (map == NULL) { beep(side); return; }
    switch (map->curtool) {
      case LOOKTOOL:
	move_look(side, map, side->ui->downx, side->ui->downy);
	break;
      case UNITMOVETOOL:
	if (map->curunit) {
	    order_moveto(map->curunit, side->ui->downx, side->ui->downy);
	} else {
	    move_look(side, map, side->ui->downx, side->ui->downy);
	}
	break;
      case UNITSHOOTTOOL:
	break;
      case HEXPAINTTOOL:
	paint_terrain(side, side->ui->downx, side->ui->downy, 0, 0,
		      map->curttype);
	break;
#ifdef LINEAR
      case BORDPAINTTOOL:
	paint_border(side,
		     side->ui->downx, side->ui->downy, side->ui->downdir,
		     map->curbordtype);
	break;
      case CONNPAINTTOOL:
	paint_connection(side,
			 side->ui->downx, side->ui->downy, side->ui->downdir,
			 map->curconntype);
	break;
#endif /* LINEAR */
      case UNITADDTOOL:
	add_unit(side, side->ui->downx, side->ui->downy,
		 map->curutype, map->curside);
	break;
      default:
	/* error eventually */
	break;
    }
#endif
}

handle_map_drag(side, map)
Side *side;
Map *map;
{	
#if 0    
    if (map == NULL) { beep(side); return; }
    switch (map->curtool) {
      case LOOKTOOL:
	move_look(side, map, side->ui->downx, side->ui->downy);
	break;
      case UNITMOVETOOL:
	if (map->curunit) {
	    /* or order a patrol? */
	    order_moveto(map->curunit, side->ui->upx, side->ui->upy);
	}
	break;
      case UNITSHOOTTOOL:
	break;
      case HEXPAINTTOOL:
	paint_terrain(side, side->ui->downx, side->ui->downy,
		      abs(side->ui->upx - side->ui->downx),
		      abs(side->ui->upy - side->ui->downy),
		      map->curttype);
	break;
#ifdef LINEAR
      case BORDPAINTTOOL:
	paint_border_line(side,
			  side->ui->downx, side->ui->downy,
			  side->ui->upx, side->ui->upy,
			  map->curbordtype);
	break;
      case CONNPAINTTOOL:
	paint_connection_line(side,
			      side->ui->downx, side->ui->downy,
			      side->ui->upx, side->ui->upy,
			      map->curconntype);
	break;
#endif /* LINEAR */
      case UNITADDTOOL:
	add_unit(side, side->ui->downx, side->ui->downy,
		 map->curutype, map->curside);
	break;
      default:
	/* error eventually */
	break;
    }
#endif
}

/* Decide what to do with a button hit on a list window. */

button_down_list(side, evt)
Side *side;
XEvent evt;
{
    int rawy, unitn;
    List *list;

    for_all_lists(side, list) {
	if (evt.xbutton.window == (Window) list->listwin) {
	    rawy = evt.xbutton.y;
	    unitn = rawy / side->ui->fh;
	    if (0 /* unitn selected */) {
	    } else {
		/* (should just select the unit) */
		
	    }
	    /* should use to select the unit or open up
	       if double-clicked */
	    return TRUE;
	}
	/* do scrollbars eventually */
    }
    return FALSE;
}

/* Decide what to do with a button release on a list window. */

button_up_list(side, evt)
Side *side;
XEvent evt;
{
    List *list;

    for_all_lists(side, list) {
	if (evt.xbutton.window == (Window) list->listwin) {
	    return TRUE;
	}
	/* do scrollbars eventually */
    }
    return FALSE;
}

/* Decide what to do with a button hit on a closeup window. */

button_down_closeup(side, evt)
Side *side;
XEvent evt;
{
    Closeup *closeup;

    for_all_closeups(side, closeup) {
	if (evt.xbutton.window == (Window) closeup->closeupwin) {
	    show_closeup(side, closeup);
	    return TRUE;
	}
    }
    return FALSE;
}

/* See if the expose event is doing some or all of a map. */

/* Decide what to do with a button hit on a closeup window. */

button_up_closeup(side, evt)
Side *side;
XEvent evt;
{
    Closeup *closeup;

    for_all_closeups(side, closeup) {
	if (evt.xbutton.window == (Window) closeup->closeupwin) {
/*	    show_closeup(side, closeup);  */
	    return TRUE;
	}
    }
    return FALSE;
}

/* See if the expose event is doing some or all of a map. */

expose_map(side, evt)
Side *side;
XEvent evt;
{
}

/* See if the expose event is doing some or all of a list. */

expose_list(side, evt)
Side *side;
XEvent evt;
{
}

/* See if the expose event is doing some or all of a closeup. */

expose_closeup(side, evt)
Side *side;
XEvent evt;
{
}

button_down_docdial(side, rawx, rawy)
Side *side;
int rawx, rawy;
{
    side->unitseveraskside = !side->unitseveraskside;
}

/* Get rid of extra input events, generally because the situation has
   changed drastically and they are no longer of interest.  This routine
   has relatively few valid uses. */

flush_input(side)
Side *side;
{
#if 0
    printf("doing an input flush\n");
    if (active_display(side)) {
	XSync(sdd(side), TRUE);  
    }
#endif
}

/* Trivial abstraction - sometimes other routines need to ensure all output
   so far is actually on the screen and not being buffered up somewhere. */

flush_output(side) 
Side *side; 
{  
    if (active_display(side)) XFlush(sdd(side));  
}

#ifdef DEBUGGING
/* Set all the window backgrounds, borders, GCs, etc, usually to reflect
   inversion of foreground and background. */


reset_color_state(side)
Side *side;
{
    XGCValues values;
    unsigned int gcmask;

    XSetWindowBackground(sdd(side), side->ui->msg, side->ui->bgcolor);
    XSetWindowBackground(sdd(side), side->ui->game, side->ui->bgcolor);
    XSetWindowBorder(sdd(side), side->ui->msg, side->ui->fgcolor);
    XSetWindowBorder(sdd(side), side->ui->game, side->ui->fgcolor);

    gcmask = GCForeground | GCBackground;
    values.foreground = side->ui->fgcolor;
    values.background = side->ui->bgcolor;
    XChangeGC(sdd(side), sd(side)->gc, gcmask, &values);
    XChangeGC(sdd(side), sd(side)->textgc, gcmask, &values);
    XChangeGC(sdd(side), sd(side)->icongc, gcmask, &values);
    XChangeGC(sdd(side), sd(side)->unitgc, gcmask, &values);
    values.foreground = side->ui->bgcolor;
    values.background = side->ui->fgcolor;
    XChangeGC(sdd(side), sd(side)->invicongc, gcmask, &values);
}
#endif /* DEBUGGING */

/* Alter the size and position of a window. */

change_window(side, win, x, y, w, h)
Side *side;
Window win;
int x, y, w, h;
{
}

